@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
+# (c) Copyright IBM Corp. 2015 - 2022
 # --------------------------------------------------------------------------
 
 """
@@ -18,22 +18,10 @@ to be performed:
  * *cpo_config.py*, a local set of changes on these parameters,
  * *cpo_config_<hostname>.py*, a hostname dependent set of changes.
 
-Each of these files is searched in the *PYTHONPATH*, and the first one that is found is read.
+Each of these files is searched first in the directory where the Pyton main file is located,
+then in the current directory, and finally in the *PYTHONPATH*. The first one that is found is read.
 Final set of parameters is obtained by reading first this module, and then those listed above.
 
-This module also defines two global variables:
-
- * *LOCAL_CONTEXT*, that contains the appropriate configuration to solve a model with a local
-   installation of the CPO solver.
-   This configuration is available for solver with version number greater or equal to 12.7.0.
-
-   This context is the context by default, referenced by the global variable 'context'.
-
- * *DOCLOUD_CONTEXT*, (deprecated) that contains the configuration necessary to solve a model on DOcloud.
-
-The method :meth:`set_default` allows to set the default configuration to one that is predefined,
-or another that has been totally customized.
-
 If called as main, this module prints the actual configuration on standard output, including
 all customizations made using the mechanism described above.
 
@@ -164,7 +152,30 @@ Configuration of the `local` solving agent
 *context.solver.local.execfile*
 
     Name or full path of the CP Optimizer Interactive executable file.
-    By default, it is set to *cpoptimizer(.exe)*, which supposes that the program is visible from the system path.
+    By default, it is set to *"cpoptimizer(.exe)"*, and is searched in the file system using the strategy described below.
+
+
+Configuration of the `lib` solving agent
+----------------------------------------
+
+*context.solver.lib.libfile*
+
+    Name or full path of the CP Optimizer library file.
+    By default, it is set to *"lib_cpo_solver_\*(.dll)"*, and is searched in the file system using the strategy described below.
+
+
+Search for executable or library files
+--------------------------------------
+
+If an executable or library file is not given with its full path, it is searched in the file system in a list of
+directories that is built in this order:
+
+   * the directory where the main Python file is located,
+   * the current directory,
+   * all directories of the environment variables 'LD_LIBRARY_PATH' or 'DYLD_LIBRARY_PATH' if defined,
+   * all directories of the environment variable 'PATH'.
+
+If the file name contains an "\*", and if there are multiple files matching the pattern, the most recent is selected.
 
 
 Configuration for best performances
@@ -187,7 +198,6 @@ Obviously, this performance is won at the cost of the loss of some features that
 Detailed description
 --------------------
 """
-
 from docplex.cp.utils import *
 from docplex.cp.parameters import CpoParameters, ALL_PARAMETER_NAMES
 
@@ -200,14 +210,17 @@ import fnmatch
 
 # Check if running in a worker environment
 try:
-    import docplex.util.environment as runenv
-    IS_IN_WORKER = isinstance(runenv.get_environment(), runenv.WorkerEnvironment)
+    from docplex_wml.version import *
+    #from docplex_wml.worker.environment import WorkerEnvironment
+    #import docplex.util.environment as runenv
+    from docplex_wml import _in_ws_nb
+    if _in_ws_nb is False:
+        IS_IN_WORKER = True
+    else:
+        IS_IN_WORKER = False
 except:
     IS_IN_WORKER = False
 
-# Indicator that program is running inside a notebook
-IS_IN_NOTEBOOK = 'ipykernel' in sys.modules
-
 # CP Optimizer Interactive executable name
 CPO_EXEC_INTERACTIVE = "cpoptimizer" + (".exe" if IS_WINDOWS else "")
 
@@ -261,7 +274,7 @@ context.model = Context()
 context.model.add_source_location = True
 
 # Minimal variable name length that trigger use of shorter alias. None for no alias.
-context.model.length_for_alias = 15
+context.model.length_for_alias = None
 
 # Automatically add a name to every top-level constraint
 context.model.name_all_constraints = False
@@ -273,7 +286,7 @@ context.model.version = None
 # Name of the directory where store copy of the generated CPO files. None for no dump.
 context.model.dump_directory = None
 
-# Flag to factorize expressions used more than ones
+# Flag to indicate to the compiler to factorize expressions used more than ones
 context.model.factorize_expressions = True
 
 # Flag to generate short model output (internal)
@@ -282,7 +295,7 @@ context.model.short_output = False
 # Type of sort for model variables. Value is in {None, 'alphabetical', 'natural')
 context.model.sort_names = None
 
-# Expression cache
+# Expression cache, global to all models
 context.model.cache = Context()
 context.model.cache.size = 10000
 context.model.cache.active = True
@@ -293,6 +306,15 @@ context.model.cache.active = True
 
 context.parser = Context()
 
+# Enable CPO parser to automatically generate blackboxes for unknown functions
+context.parser.auto_blackbox = False
+
+# Indicate to print warning messages (for example for unknown blackbox functions) at the end of parsing
+context.parser.print_warnings = True
+
+# Indicate to check format version and raise an exception if out of range
+context.parser.check_format_version = True
+
 # Indicate to FZN parser to reduce model when possible
 context.parser.fzn_reduce = False
 
@@ -318,7 +340,10 @@ context.solver = Context()
 context.solver.trace_cpo = False
 
 # Indicate to trace solver log on log_output.
-context.solver.trace_log = False
+context.solver.trace_log = True
+if is_in_notebook():
+    # In notebooks, no traces by default because it may crash the environment
+    context.solver.trace_log = False
 
 # Enable undocumented parameters
 context.solver.enable_undocumented_params = False
@@ -332,8 +357,8 @@ context.solver.add_log_to_solution = True
 # Indicate to add the conflict in CPO format to conflict refiner result
 context.solver.add_conflict_as_cpo = True
 
-# Indicate to replace simple solve by a start/next loop
-context.solver.solve_with_start_next = False
+# Indicate to replace simple solve by a search_next loop
+context.solver.solve_with_search_next = False
 
 # Log prefix
 context.solver.log_prefix = "[Solver] "
@@ -389,7 +414,6 @@ context.solver.local.log_prefix = "[Local] "
 # Local sub-process start timeout in seconds
 context.solver.local.process_start_timeout = 5
 
-
 #-----------------------------------------------------------------------------
 # Local solving with CPO library (internal)
 
@@ -417,69 +441,55 @@ except:
 
 
 #-----------------------------------------------------------------------------
-# DoCloud solving agent context (deprecated)
+# Solver simulators
 
-context.solver.docloud = Context()
+# Solver simulator that always fails
+context.solver.simulatorfail = Context()
+context.solver.simulatorfail.class_name = "docplex.cp.solver.solver_simulator.CpoSolverSimulatorFail"
+context.solver.simulatorfail.log_prefix = "[SimulatorFail] "
 
-# Python class implementing the agent
-context.solver.docloud.class_name = "docplex.cp.solver.solver_docloud.CpoSolverDocloud"
+# Solver simulator generating a random solution
+context.solver.simulatorrandom = Context()
+context.solver.class_name = "docplex.cp.solver.solver_simulator.CpoSolverSimulatorRandom"
+context.solver.log_prefix = "[SimulatorRandom] "
 
-# Url of the DOCloud service
-context.solver.docloud.url = "https://api-oaas.docloud.ibmcloud.com/job_manager/rest/v1/"
 
-# Authentication key.
-context.solver.docloud.key = "'Set your key in docloud_config.py''"
+#-----------------------------------------------------------------------------
+# Parameters for solver interactive monitor
 
-# Secret key.
-context.solver.docloud.secret = None
+context.interactive = Context()
 
-# Indicate to verify SSL certificates
-context.solver.docloud.verify_ssl = True
+# Confirm solver monitor window closing
+context.interactive.user_preferences_file = "DocplexCpUserPreferences.prf"
 
-# Proxies (map protocol_name/endpoint, as described in http://docs.python-requests.org/en/master/user/advanced/#proxies)
-context.solver.docloud.proxies = None
+# Confirm solver monitor window closing
+context.interactive.window_confirm_exit = True
 
-# Default unitary request timeout in seconds
-context.solver.docloud.request_timeout = 30
+# Abort solve when closing window
+context.interactive.solver_abort_on_exit = True
 
-# Time added to expected solve time to compute the total result waiting timeout
-context.solver.docloud.result_wait_extra_time = 60
-
-# Clean job after solve indicator
-context.solver.docloud.clean_job_after_solve = True
-
-# Add 'Connection close' in all headers
-context.solver.docloud.always_close_connection = False
-
-# Log prefix
-context.solver.docloud.log_prefix = "[DOcloud] "
-
-# Polling delay (min, max and increment)
-context.solver.docloud.polling = Context(min=1, max=3, incr=0.2)
+# Daemon display thread (avoid tk exception if listener)
+context.interactive.daemon_thread = True
 
 
 #-----------------------------------------------------------------------------
 # Apply special changes if running in a worker
 
 if IS_IN_WORKER:
-    context.solver.max_threads = runenv.get_environment().get_available_core_count()
+    try:
+        context.solver.max_threads = int(os.environ["CORE_NUMBERS"])
+    except:
+        context.solver.max_threads = 1
+    #context.solver.max_threads = runenv.get_environment().get_available_core_count()
 
 
 #-----------------------------------------------------------------------------
-# Create special context for docloud
+# Create context variable
 
-# Create 2 contexts for local and docloud
+# Create a local context variable
+# Now useless, comes from a time where there was LOCAL_CONTEXT and DOCLOUD_CONTEXT.
+# Kept for ascending compatibility
 LOCAL_CONTEXT = context
-DOCLOUD_CONTEXT = context.clone()
-
-# Set local context specific attributes
-LOCAL_CONTEXT.solver.trace_log = not IS_IN_NOTEBOOK
-LOCAL_CONTEXT.model.length_for_alias = None
-
-# Set DOcloud context specific attributes
-DOCLOUD_CONTEXT.solver.agent = 'docloud'
-DOCLOUD_CONTEXT.solver.trace_log = False
-DOCLOUD_CONTEXT.model.length_for_alias = 15
 
 
 #=============================================================================
@@ -524,7 +534,7 @@ def _get_port_name():
     """
     sstm = platform.system()
     if sstm == 'Windows': return 'x64_win64'
-    if sstm == 'Darwin':  return 'x86-64osx'
+    if sstm == 'Darwin':  return 'x86-64_osx'
     if sstm == 'Linux':
         machine = platform.machine()
         if machine == 'x86_64':
@@ -543,28 +553,42 @@ def _build_search_path (ctx):
     Args:
         ctx:  Context to get information from
     """
-    # Initialize with docplex install directory
-    python_home = os.path.dirname(os.path.abspath(sys.executable))
-    if IS_WINDOWS:
-        path = [os.path.join(python_home, "Scripts")]
-        appdata = os.environ.get('APPDATA')
-        if appdata is not None:
-            path.append(os.path.join(appdata, os.path.join('Python', 'Scripts')))
-    else:
-        path = ["~/.local/bin", os.path.join(python_home, "bin")]
+    # Initialize search path
+    path = []
+
+    # Set it to COS location if provided
+    cosloc = ctx.get_by_path('cos.location')
+    if cosloc:
+        # Check existence of the directory
+        if not os.path.isdir(cosloc):
+            raise Exception("COS location directory '{}' does not exists.".format(cosloc))
+        # As explicitly defined, add it in front of search path
+        path.append(cosloc + "/bin/" + _get_port_name())
+
+    # Add directory where python main is located
+    mpfile = get_main_file()
+    if mpfile:
+        path.append(os.path.dirname(mpfile))
+
+    # Add current directory
+    path.append('.')
 
     # Add all system path
     path.extend(get_system_path())
 
-    # Add COS location if defined
-    cdir = ctx.get_by_path('cos.location')
-    if cdir:
-        # Check existence of the directory
-        if not os.path.isdir(cdir):
-            raise Exception("COS location directory '{}' does not exists.".format(cdir))
-        # As explicitly defined, add it in front of search path
-        path = [cdir + "/bin/" + _get_port_name()] + path
+    # Append docplex install directory
+    python_home = os.path.dirname(os.path.abspath(sys.executable))
+    if IS_WINDOWS:
+        path.append(os.path.join(python_home, "Scripts"))
+        appdata = os.environ.get('APPDATA')
+        if appdata is not None:
+            path.append(os.path.join(appdata, os.path.join('Python', 'Scripts')))
     else:
+        path.append("~/.local/bin")
+        path.append(os.path.join(python_home, "bin"))
+
+    # Add all COS locations if not already forced
+    if not cosloc:
         # Add all existing installed COS starting from the most recent
         for cnv in _COS_ROOT_DIRS_ENV_VARS:
             cdir = os.getenv(cnv)
@@ -581,7 +605,7 @@ def _search_exec_file(file, ctx):
         file:  Executable file name
         ctx:   Context to get information from
     Returns:
-        Full path of the first executable file found, None if not found
+        Full path of the first executable file found, given file if not found
     """
     # Check null
     if not file:
@@ -590,24 +614,28 @@ def _search_exec_file(file, ctx):
     if is_exe_file(file):
         return file
     # Check if file contains a path
-    if os.path.basename(file) == file:
+    fname = os.path.basename(file)
+    if fname == file:
         # Build full search path
         path = _build_search_path(ctx)
     else:
         # Keep file path as single path
         path = [os.path.dirname(file)]
-        file = os.path.basename(file)
     # Check if file contains a pattern
-    if "*" in file:
+    if "*" in fname:
         # Check in the path
         for d in path:
             if not os.path.isdir(d):
                 continue
-            lf = [os.path.join(d, f) for f in os.listdir(d) if fnmatch.fnmatch(f, file)]
+            lf = [os.path.join(d, f) for f in os.listdir(d) if fnmatch.fnmatch(f, fname)]
             if lf:
                 # Take most recent executable file
                 lf = [f for f in lf if is_exe_file(f)]
                 if lf:
+                    # Skip "ai" files if any
+                    nlf = [f for f in lf if not os.path.splitext(f)[0].endswith("ai")]
+                    if nlf:
+                        lf = nlf
                     lf.sort(key=lambda x: os.path.getmtime(x))
                     return lf[-1]
     else:
@@ -615,11 +643,11 @@ def _search_exec_file(file, ctx):
         for d in path:
             if not os.path.isdir(d):
                 continue
-            nf = os.path.join(d, file)
+            nf = os.path.join(d, fname)
             if is_exe_file(nf):
                 return nf
 
-    return None
+    return file
 
 
 # Attribute values denoting a default value
@@ -650,13 +678,6 @@ def _get_effective_context(**kwargs):
     Returns:
         Updated (cloned) context
     """
-    # If 'url' and 'key' are defined, force agent to be docloud
-    if ('agent' not in kwargs) and not IS_IN_WORKER:
-        url = kwargs.get('url')
-        key = kwargs.get('key')
-        if url and key and is_string(url) and is_string(key) and url.startswith('http'):
-            kwargs['agent'] = 'docloud'
-
     # Determine source context
     ctx = kwargs.get('context')
     if (ctx is None) or (ctx in DEFAULT_VALUES):
@@ -668,13 +689,16 @@ def _get_effective_context(**kwargs):
     # First set parameters if given
     prms = kwargs.get('params')
     if prms is not None:
-        ctx.params.add(prms)
+        ctx.params.set_other(prms)
 
     # Process other changes, check first if undocumented params are enabled (or not)
     uk = 'enable_undocumented_params'
     if uk in kwargs:
         _change_context_attribute(ctx, uk, kwargs[uk])
     for k, v in kwargs.items():
+        # Replace obsolete solve_with_start_next attribute
+        if k == 'solve_with_start_next':
+            k = 'solve_with_search_next'
         if (k != 'context') and (k != 'params') and (k != uk) and (v not in DEFAULT_VALUES):
             _change_context_attribute(ctx, k, v)
 
@@ -695,14 +719,24 @@ def _get_effective_context(**kwargs):
 # Overload default context with optional customizations
 #=============================================================================
 
+# Compute config files search path
+_CONFIG_SEARCH_PATH = []
+mpfile = get_main_file()     # Add directory of the main file if any
+if mpfile:
+    _CONFIG_SEARCH_PATH.append(os.path.dirname(mpfile))
+_CONFIG_SEARCH_PATH.append('.')       # Add current directory
+_CONFIG_SEARCH_PATH.extend(sys.path)  # Add all PYTHONPATH
+
+
 def _eval_file(file):
     """ If found in python path, evaluate the content of a python module in this module.
     Args:
         file: Python file to evaluate
     """
-    for dir in sys.path:
+    for dir in _CONFIG_SEARCH_PATH:
         f = dir + "/" + file if dir else file
         if os.path.isfile(f):
+            # print("Loading configuration file '{}'".format(f))
             try:
                 with open(f) as fin:
                     fcont = fin.read()
@@ -714,9 +748,14 @@ def _eval_file(file):
                 raise Exception("Error while loading config file {}: {}".format(f, str(e)))
 
 # Load all config changes
-for f in ("cpo_config.py", "cpo_config_" + socket.gethostname() + ".py", "docloud_config.py"):
+for f in ("cpo_config.py", "cpo_config_" + socket.gethostname() + ".py",):
     _eval_file(f)
 
+# Check particular case of renamed solve_with_start_next
+ov = context.solver.solve_with_start_next
+if ov is not None:
+    context.solver.solve_with_search_next = ov
+
 # Overwrite with environment definitions if any
 envchg = os.environ.get(CONTEXT_ENVIRONMENT)
 if envchg:
