@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 # Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis
 
@@ -17,6 +17,7 @@ from docplex.cp.catalog import *
 from docplex.cp.solution import *
 from docplex.cp.model import CpoModel
 import docplex.cp.config as config
+from docplex.cp.blackbox import CpoBlackboxFunction, CpoBlackboxFunctionCall
 import math
 import traceback
 
@@ -29,7 +30,7 @@ import traceback
 MIN_CPO_VERSION_NUMBER = "12.6.0.0"
 
 # Maximum CPO format version number
-MAX_CPO_VERSION_NUMBER = "12.10.0.0"
+MAX_CPO_VERSION_NUMBER = "21.10.0.0"
 
 # Map of all operators. Key is operator, value is list of corresponding operation descriptors
 _ALL_OPERATORS = {}
@@ -42,8 +43,6 @@ for op in ALL_OPERATIONS:
     if op.priority >= 0:
         _ALL_OPERATORS[op.keyword] = op
     _ALL_OPERATIONS[op.cpo_name] = op
-_ALL_OPERATIONS['alldiff'] = Oper_all_diff
-_ALL_OPERATIONS['allDiff'] = Oper_all_diff
 
 # Known identifiers
 _KNOWN_IDENTIFIERS = {"intmax": INT_MAX, "intmin": INT_MIN,
@@ -57,6 +56,7 @@ _ARRAY_TYPES = {'intArray': Type_IntArray, 'floatArray': Type_FloatArray,
                 'intervalVarArray': Type_IntervalVarArray, 'sequenceVarArray': Type_SequenceVarArray,
                 'intValueSelectorArray': Type_IntValueSelectorArray, 'intVarSelectorArray': Type_IntVarSelectorArray,
                 'tupleSet' : Type_TupleSet,
+                'cumulExprArray': Type_CumulExprArray,
                 '_cumulAtomArray' : Type_CumulAtomArray}
 
 # Fake operator '..' to read intervals properly
@@ -72,7 +72,7 @@ class CpoParserException(CpoException):
     """ The base class for exceptions raised by the CPO parser
     """
     def __init__(self, msg):
-        """ Create a new exception
+        """ **Constructor**
         Args:
             msg: Error message
         """
@@ -83,7 +83,7 @@ class CpoUnsupportedFormatVersionException(CpoParserException):
     """ Exception raised when a version mismatch is detected
     """
     def __init__(self, msg):
-        """ Create a new exception
+        """ **Constructor**
         Args:
             msg: Error message
         """
@@ -101,13 +101,16 @@ class CpoParser(object):
                  'pushtoken',      # Pushed token
                  'fun_handlers',   # Special function handlers
                  'current_loc',    # Current source location
+                 'context',        # Parser configuration context
+                 'auto_blackbox',  # Auto generate blackbox functions for unknown functions
                  )
-    
-    def __init__(self, mdl=None):
-        """ Create a new CPO format parser
+
+    def __init__(self, mdl=None, **kwargs):
+        """ **Constructor**
 
         Args:
-            mdl:  Model to fill, None (default) to create a new one.
+            mdl:       Model to fill, None (default) to create a new one.
+            **kwargs: (Optional) changes to context
         """
         super(CpoParser, self).__init__()
         self.model = mdl if mdl is not None else CpoModel()
@@ -117,6 +120,8 @@ class CpoParser(object):
         self.token = None
         self.pushtoken = None
         self.current_loc = None
+        self.context = config._get_effective_context(**kwargs).parser
+        self.auto_blackbox = self.context.auto_blackbox
 
         # Do not store location information (would store parser instead of real lines)
         self.model.source_loc = False
@@ -135,6 +140,7 @@ class CpoParser(object):
             'stepFunction':      self._read_fun_stepFunction,
             'segmentedFunction': self._read_fun_segmentedFunction,
             'transitionMatrix':  self._read_fun_transitionMatrix,
+            '_intervalArray':    self._read_fun_intervalArray,
         }
 
         # TODO: parse and include source information ?
@@ -170,6 +176,8 @@ class CpoParser(object):
         self.tokenizer.close()
         self.tokenizer = None
 
+        self._final_checkings()
+
         return self.model
 
 
@@ -187,6 +195,9 @@ class CpoParser(object):
         self._read_statement_list()
         self.tokenizer.close()
         self.tokenizer = None
+
+        self._final_checkings()
+
         return self.model
 
 
@@ -289,15 +300,12 @@ class CpoParser(object):
             self._raise_exception("Line number should be an integer")
         lnum = int(tok.value)
         # Get optional source file name, string on the same line
-        cline = self.tokenizer.line_number
-        tok = self._next_token()
-        if cline != self.tokenizer.line_number:
-            self._push_token(tok)
-            fname = None if self.current_loc is None else self.current_loc[0]
-        else:
-            if tok.type != TOKEN_TYPE_STRING:
-               self._raise_exception("File name should be a string")
-            fname = tok.get_string()
+        # String is read not as a token, because file pathes may contain backslashes on windows.
+        fname = self.tokenizer._get_line_reminder().strip()
+        if fname.startswith('"'):
+            fname = fname[1:-1]
+        if not fname:
+            fname = None
         self.current_loc = (fname, lnum)
 
 
@@ -437,7 +445,7 @@ class CpoParser(object):
         Return:
             Expression that has been read
         """
-
+        auto_blackbox = self.auto_blackbox
         tok = self.token
         toktyp = tok.type
         ntok = self._next_token()
@@ -445,7 +453,7 @@ class CpoParser(object):
         # Check int constant
         if toktyp is TOKEN_TYPE_INTEGER:
             return int(tok.value)
-        
+
         # Check float constant
         if toktyp is TOKEN_TYPE_FLOAT:
             return float(tok.value)
@@ -457,7 +465,7 @@ class CpoParser(object):
             # Read next expression
             expr = self._read_sub_expression()
             return self._create_operation_expression(op, (expr,))
-        
+
         # Check symbol
         if toktyp is TOKEN_TYPE_SYMBOL:
             tokval = tok.value
@@ -467,13 +475,41 @@ class CpoParser(object):
                 fun = self.fun_handlers.get(tokval)
                 if fun:
                     return fun()
-
-                # General function call, retrieve operation descriptor
-                op = _ALL_OPERATIONS.get(tokval)
-                if op is None:
-                    self._raise_exception("Unknown operation '" + str(tok.value) + "'")
+                # General function call, read arguments
                 args = self._read_expression_list_up_to_parent_close()
-                return self._create_operation_expression(op, args)
+                # Retrieve operation descriptor
+                op = _ALL_OPERATIONS.get(tokval)
+                if op is not None:
+                    # Check blackbox evaluation function
+                    if auto_blackbox and (op is Oper_eval):
+                        # Update bbf dimension
+                        assert len(args) == 2
+                        bbfc, vx = args
+                        assert isinstance(bbfc, CpoBlackboxFunctionCall)
+                        assert is_int(vx)
+                        bbfc.blackbox._update_dimension(vx)
+                    # Build operation call
+                    return self._create_operation_expression(op, args)
+                # Function call, build argument expressions
+                argexprs = tuple(map(build_cpo_expr, args))
+                # Search in blackbox functions
+                bbf = self.model.get_blackbox_function(tokval)
+                if bbf is not None:
+                    # if auto-created, update argument types
+                    if bbf.auto:
+                        bbf._update_check_arg_types(argexprs)
+                    return CpoBlackboxFunctionCall(bbf, argexprs)
+                # Check if auto blackbox creation is enabled
+                if auto_blackbox:
+                    bbf = CpoBlackboxFunction(name=tokval, dimension=None)
+                    obbf = self.model.get_blackbox_function(tokval)
+                    if obbf is None:
+                        bbf._update_check_arg_types(argexprs)
+                        self.model.add_blackbox_function(bbf)
+                    else:
+                        obbf._update_check_arg_types(argexprs)
+                    return CpoBlackboxFunctionCall(bbf, argexprs)
+                self._raise_exception("Unknown operation '{}'".format(tokval))
 
             if ntok is TOKEN_HOOK_OPEN:
                 # Read typed array
@@ -498,7 +534,7 @@ class CpoParser(object):
 
             # Token is an expression id
             return self._get_identifier_value(tok.get_string())
-        
+
         # Check expression in parenthesis
         if tok is TOKEN_PARENT_OPEN:
             expr = self._read_expression_list_up_to_parent_close()
@@ -510,11 +546,11 @@ class CpoParser(object):
         if tok is TOKEN_HOOK_OPEN:
             expr = self._read_expression_list(TOKEN_HOOK_CLOSE)
             return expr
-            
+
         # Check reference to a model expression or variable
         if toktyp is TOKEN_TYPE_STRING:
             return self._get_identifier_value(tok.get_string())
-                
+
         # Unknown expression
         self._raise_exception("Invalid start of expression: '" + str(tok) + "'")
 
@@ -622,6 +658,15 @@ class CpoParser(object):
         return CpoTransitionMatrix(values=(mtrx[i * size : (i+1) * size] for i in range(size)))
 
 
+    def _read_fun_intervalArray(self):
+        """ Read a function call to '_intervalArray'
+        Returns: New expression
+        """
+        # Read list of couples
+        lcpls = self._read_expression_list_up_to_parent_close()
+        return CpoValue(lcpls, Type_IntervalArray)
+
+
     def _read_expression_list_in_parenthesis(self):
         """ Read a list of expressions between parenthesis
         Opening parenthesis is read and checked by this method.
@@ -757,12 +802,13 @@ class CpoParser(object):
                 if vtok.type is TOKEN_TYPE_VERSION:
                     ver = vtok.get_string()
                     self.model.set_format_version(ver)
-                    if compare_natural(ver, MIN_CPO_VERSION_NUMBER) < 0:
-                        raise CpoUnsupportedFormatVersionException("Can not parse a CPO file with version {}, lower than {}"
-                                                                   .format(ver, MIN_CPO_VERSION_NUMBER))
-                    if compare_natural(ver, MAX_CPO_VERSION_NUMBER) > 0:
-                        raise CpoUnsupportedFormatVersionException("Can not parse a CPO file with version {}, greater than {}"
-                                                                   .format(ver, MAX_CPO_VERSION_NUMBER))
+                    if self.context.check_format_version:
+                        if compare_natural(ver, MIN_CPO_VERSION_NUMBER) < 0:
+                            raise CpoUnsupportedFormatVersionException("Can not parse a CPO file with version {}, lower than {}"
+                                                                       .format(ver, MIN_CPO_VERSION_NUMBER))
+                        if compare_natural(ver, MAX_CPO_VERSION_NUMBER) > 0:
+                            raise CpoUnsupportedFormatVersionException("Can not parse a CPO file with version {}, greater than {}"
+                                                                       .format(ver, MAX_CPO_VERSION_NUMBER))
                 self._check_token(self._next_token(), TOKEN_PARENT_CLOSE)
             tok = self._next_token()
 
@@ -957,7 +1003,7 @@ class CpoParser(object):
         Args:
             tok: Token
         Returns:
-            String value of the token            
+            String value of the token
         """
         if tok.type is TOKEN_TYPE_SYMBOL:
             return tok.value
@@ -1025,7 +1071,7 @@ class CpoParser(object):
         Args:
             tok:  Unexpected token
         """
-        if tok is None: 
+        if tok is None:
             tok = self.token
         if expect is None:
             self._raise_exception("Unexpected token '" + str(tok) + "'")
@@ -1059,8 +1105,25 @@ class CpoParser(object):
     def _push_token(self, tok):
         """ Push current token
         Args:
-            tok: New current token 
+            tok: New current token
         """
         self.pushtoken = self.token
         self.token = tok
-        
+
+
+    def _final_checkings(self):
+        """ Proceed to final model checkings
+        """
+        # Check automatic blackboxes never used (no eval)
+        for bbf in self.model.get_blackbox_functions():
+            if bbf.auto and bbf.dimension is None:
+                raise CpoParserException("Function '{}' does not exists as modeling function is is not used as blackbox.".format(bbf.get_name()))
+
+        # Skip if warnings not enabled
+        if not self.context.print_warnings:
+            return
+
+        # Check if all BBF functions have an implementation
+        for bbf in self.model.get_blackbox_functions():
+            if bbf.get_implementation() is None:
+                print("WARNING: Blackbox function '{}' does not have an implementation.".format(bbf))
