@@ -1663,7 +1663,7 @@ class FznParser(object):
         """ Constrains the elements of x to define a subcircuit where x[i] = j means that j is the successor of i and x[i] = i means that i is not in the circuit.
         """
         x = [0] + list(self._get_cpo_expr(fc.args[0]).value)
-        self._add_to_model(CpoFunctionCall(Oper__sub_circuit, Type_Constraint, (build_cpo_expr(x),) ))
+        self._add_to_model(CpoFunctionCall(Oper_sub_circuit, Type_Constraint, (build_cpo_expr(x),) ))
 
 
     def _compile_pred_inverse(self, fc):
@@ -1825,8 +1825,8 @@ class FznParser(object):
         """
         dom = var.domain
         #print("   constrain expression {} to domain {}".format(expr, dom))
-        dmin = expression._domain_min(dom)
-        dmax = expression._domain_max(dom)
+        dmin = expression.get_domain_min(dom)
+        dmax = expression.get_domain_max(dom)
 
         # Check boolean expression
         if dmin == 0 and dmax == 1:
