@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2016, 2017, 2018
+# (c) Copyright IBM Corp. 2016 .. 2022
 # --------------------------------------------------------------------------
 # Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis
 
@@ -13,31 +13,41 @@ a local CP Optimizer Interactive (cpoptimizer(.exe)).
 from docplex.cp.solution import *
 from docplex.cp.utils import CpoException
 from docplex.cp.solver.solver import CpoSolver, CpoSolverAgent, CpoSolverException
+from docplex.cp.catalog import *
+from docplex.cp.blackbox import BLACKBOX_ARGUMENT_TYPES_ENCODING
 
 import subprocess
 import sys
 import time
 import threading
 import json
-import signal, os
+import os
+import traceback
 
 
-###############################################################################
-##  Private constants
-###############################################################################
+#-----------------------------------------------------------------------------
+#  Private constants
+#-----------------------------------------------------------------------------
+
+# Version of this client
+#CLIENT_VERSION = 8 # last OO's version
+CLIENT_VERSION = 9
 
 # List of command ids that can be sent to solver
-CMD_EXIT            = "Exit"            # End process (no data)
-CMD_SET_CPO_MODEL   = "SetCpoModel"     # CPO model as string
-CMD_SOLVE_MODEL     = "SolveModel"      # Complete solve of the model (no data)
-CMD_START_SEARCH    = "StartSearch"     # Start search (no data)
-CMD_SEARCH_NEXT     = "SearchNext"      # Get next solution (no data)
-CMD_END_SEARCH      = "EndSearch"       # End search (no data)
-CMD_REFINE_CONFLICT = "RefineConflict"  # Refine conflict (no data)
-CMD_PROPAGATE       = "Propagate"       # Propagate (no data)
-CMD_RUN_SEEDS       = "RunSeeds"        # Run with multiple seeds.
-CMD_ADD_CALLBACK    = "AddCallback"     # Add callback proxy to the solver
-MD_SET_FAILURE_TAGS = "SetFailureTags"  # Give list of failure tags to explain.
+CMD_EXIT                = "Exit"              # End process (no data)
+CMD_SET_CPO_MODEL       = "SetCpoModel"       # CPO model as string
+CMD_SOLVE_MODEL         = "SolveModel"        # Complete solve of the model (no data)
+CMD_START_SEARCH        = "StartSearch"       # Start search (no data)
+CMD_SEARCH_NEXT         = "SearchNext"        # Get next solution (no data)
+CMD_END_SEARCH          = "EndSearch"         # End search (no data)
+CMD_ABORT_SEARCH        = "AbortSearch"       # Abort search (no data)
+CMD_REFINE_CONFLICT     = "RefineConflict"    # Refine conflict (no data)
+CMD_PROPAGATE           = "Propagate"         # Propagate (no data)
+CMD_RUN_SEEDS           = "RunSeeds"          # Run with multiple seeds.
+CMD_ADD_CALLBACK        = "AddCallback"       # Add callback proxy to the solver
+CMD_SET_FAILURE_TAGS    = "SetFailureTags"    # Give list of failure tags to explain.
+CMD_ADD_BLACKBOX_FUN    = "AddBlackboxFun"    # Add/register a blackbox function
+CMD_SET_CLIENT_VERSION  = "SetClientVersion"  # Set the version number of the client
 
 # List of events received from solver
 EVT_VERSION_INFO        = "VersionInfo"        # Local solver version info (String in JSON format)
@@ -54,6 +64,10 @@ EVT_PROPAGATE_RESULT    = "PropagateResult"    # Propagate result in JSON format
 EVT_RUN_SEEDS_RESULT    = "RunSeedsResult"     # Run seeds result (no data, all is in log)
 EVT_CALLBACK_EVENT      = "CallbackEvent"      # Callback event. Data is event name.
 EVT_CALLBACK_DATA       = "CallbackData"       # Callback data, following event. Data is JSON document.
+EVT_BLACKBOX_FUNCALL    = "BlackboxFunCall"    # Call to a blackbox function for evaluation.
+
+# Events sent back to solver
+EVT_BLACKBOX_RESULT     = "BlackboxResult"     # Result of a blackbox function call
 
 # Max possible received data size in one message
 _MAX_RECEIVED_DATA_SIZE = 1000000
@@ -61,13 +75,10 @@ _MAX_RECEIVED_DATA_SIZE = 1000000
 # Python 2 indicator
 IS_PYTHON_2 = (sys.version_info[0] == 2)
 
-# Version of this client
-CLIENT_VERSION = 4
 
-
-###############################################################################
-##  Public classes
-###############################################################################
+#-----------------------------------------------------------------------------
+#  Public classes
+#-----------------------------------------------------------------------------
 
 class CpoSolverLocal(CpoSolverAgent):
     """ Interface to a local solver through an external process """
@@ -78,14 +89,16 @@ class CpoSolverLocal(CpoSolverAgent):
                  'available_commands',  # List of available commands
                  'timeout_kill',        # Indicates process have been killed by timeout
                  'out_lock',            # Lock to protect output stream
+                 'proxy_version',       # Version of the remote proxy
+                 'os',                  # Hack for Python 38 to avoid display of "OSError: [WinError 6] The handle is invalid"
+                                        # when this object is deleted. (Don't ask why this work, it is a mystery)
                 )
 
-    def __init__(self, solver, params, context):
+    def __init__(self, solver, context):
         """ Create a new solver that solves locally with CP Optimizer Interactive.
 
         Args:
             solver:  Parent solver
-            params:  Solving parameters
             context: Solver context
         Raises:
             CpoException if proxy executable does not exists
@@ -94,15 +107,19 @@ class CpoSolverLocal(CpoSolverAgent):
         self.process = None
         self.active = True
         self.timeout_kill = False
-        super(CpoSolverLocal, self).__init__(solver, params, context)
+        super(CpoSolverLocal, self).__init__(solver, context)
+        self.process_infos['ClientVersion'] = CLIENT_VERSION
 
         # Check if executable file exists
         xfile = context.execfile
         if xfile is None or not is_string(xfile):
+            self._notify_execfile_not_found()
             raise CpoException("Executable file should be given in 'execfile' context attribute.")
         if not os.path.isfile(xfile):
+            self._notify_execfile_not_found()
             raise CpoException("Executable file '{}' does not exists".format(xfile))
         if not is_exe_file(xfile):
+            self._notify_execfile_not_found()
             raise CpoException("Executable file '{}' is not executable".format(xfile))
 
         # Create solving process
@@ -113,19 +130,21 @@ class CpoSolverLocal(CpoSolverAgent):
         try:
             self.process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, universal_newlines=False)
         except:
+            self._notify_execfile_not_found()
             raise CpoException("Can not execute command '{}'. Please check availability of required executable file.".format(' '.join(cmd)))
         self.pout = self.process.stdin
         self.pin = self.process.stdout
         self.out_lock = threading.Lock()
 
         # Read initial version info from process
-        self.version_info = None
         timeout = context.process_start_timeout
         timer = threading.Timer(timeout, self._process_start_timeout)
         timer.start()
         try:
             evt, data = self._read_message()
         except Exception as e:
+            if context.log_exceptions:
+                traceback.print_exc()
             if self.timeout_kill:
                 raise CpoSolverException("Solver process was too long to start and respond ({} seconds). Process has been killed.".format(timeout))
             raise CpoSolverException("Solver sub-process start failure: {}".format(e))
@@ -136,14 +155,14 @@ class CpoSolverLocal(CpoSolverAgent):
             raise CpoSolverException("Unexpected event {} received instead of version info event {}.".format(evt, EVT_VERSION_INFO))
         self.version_info = verinf = json.loads(data)
         self.available_commands = self.version_info['AvailableCommands']
+        self.proxy_version = self.version_info.get('ProxyVersion', 0)
         # Normalize information
         verinf['AgentModule'] = __name__
 
         context.log(3, "Local solver info: '", verinf, "'")
 
-        # Transfer infos in process info
-        for x in ('ProxyVersion', 'AngelVersion', 'SourceDate', 'SolverVersion'):
-            self.process_infos[x] = self.version_info.get(x)
+        # Transfer all solver infos in process info
+        self.process_infos.update(self.version_info)
 
         # Check solver version if any
         sver = self.version_info.get('SolverVersion')
@@ -151,6 +170,14 @@ class CpoSolverLocal(CpoSolverAgent):
         if sver and mver and compare_natural(mver, sver) > 0:
             raise CpoSolverException("Solver version {} is lower than model format version {}.".format(sver, mver))
 
+        # Send client version if accepted by server
+        avcmds = verinf.get('AvailableCommands')
+        if (avcmds is not None) and (CMD_SET_CLIENT_VERSION in avcmds):
+            nbfrm = bytearray(4)
+            encode_integer_big_endian_4(CLIENT_VERSION, nbfrm, 0)
+            self._write_message(CMD_SET_CLIENT_VERSION, data=nbfrm)
+            self._wait_event(EVT_SUCCESS)
+
 
     def _process_start_timeout(self):
         """ Process the raise of start timeout timer """
@@ -161,11 +188,6 @@ class CpoSolverLocal(CpoSolverAgent):
             self.process.kill()
 
 
-    def __del__(self):
-        # End solve
-        self.end()
-
-
     def solve(self):
         """ Solve the model
 
@@ -202,12 +224,14 @@ class CpoSolverLocal(CpoSolverAgent):
 
         self._write_message(CMD_START_SEARCH)
 
+        # Wait for answer message, only for version > 14 (for compatibility)
+        if self.proxy_version >= 14:
+            self._wait_event(EVT_SUCCESS)
+
 
     def search_next(self):
         """ Get the next available solution.
 
-        (This method starts search automatically.)
-
         Returns:
             Next model result (type CpoSolveResult)
         """
@@ -243,10 +267,17 @@ class CpoSolverLocal(CpoSolverAgent):
         """ Abort current search.
         This method is designed to be called by a different thread than the one currently solving.
         """
-        try:
-            self.process.kill()
-        except:
-            pass
+        # Check if remote angel is capable of processing abort_search
+        if self._is_abort_search_supported():
+            # Request abort search
+            self._write_message(CMD_ABORT_SEARCH)
+            # No event returned back because solve is supposed to wait for another
+        else:
+            # Old fashion, just kill the sub-process
+            try:
+                self.process.kill()
+            except:
+                pass
 
 
     def refine_conflict(self):
@@ -268,19 +299,19 @@ class CpoSolverLocal(CpoSolverAgent):
             self._write_message(CMD_REFINE_CONFLICT, bytearray([1]))
             # Wait JSON result
             jsol = self._wait_json_result(EVT_CONFLICT_RESULT)
-            # Wait for CPO conflict
-            cposol = self._wait_event(EVT_CONFLICT_RESULT_CPO)
+            result = self._create_result_object(CpoRefineConflictResult, jsol)
+            if result.is_conflict():
+                # Wait for CPO conflict
+                cposol = self._wait_event(EVT_CONFLICT_RESULT_CPO)
+                result.cpo_conflict = cposol
         else:
             # Request refine conflict
             self._write_message(CMD_REFINE_CONFLICT)
             # Wait JSON result
             jsol = self._wait_json_result(EVT_CONFLICT_RESULT)
-            # No CPO conflict
-            cposol = None
+            result = self._create_result_object(CpoRefineConflictResult, jsol)
 
-        # Build result object
-        result = self._create_result_object(CpoRefineConflictResult, jsol)
-        result.cpo_conflict = cposol
+        # Return result object
         return result
 
 
@@ -358,7 +389,7 @@ class CpoSolverLocal(CpoSolverAgent):
         self._init_model_in_solver()
 
         # Check command availability
-        if MD_SET_FAILURE_TAGS not in self.available_commands:
+        if CMD_SET_FAILURE_TAGS not in self.available_commands:
             raise CpoNotSupportedException("Method 'set_explain_failure_tags' is not available in local solver '{}'".format(self.context.execfile))
 
         # Build list of tags
@@ -373,7 +404,7 @@ class CpoSolverLocal(CpoSolverAgent):
             encode_integer_big_endian_4(t, tagfrm, 4 * (i + 1))
 
         # Send command
-        self._write_message(MD_SET_FAILURE_TAGS, data=tagfrm)
+        self._write_message(CMD_SET_FAILURE_TAGS, data=tagfrm)
         self._wait_event(EVT_SUCCESS)
 
 
@@ -383,29 +414,39 @@ class CpoSolverLocal(CpoSolverAgent):
         if self.active:
             self.active = False
             try:
+                #traceback.print_stack()
                 self._write_message(CMD_EXIT)
-            except:
-                pass
-            try:
-                self.pout.close()
-            except:
-                pass
-            try:
-                self.pin.close()
-            except:
+            except BaseException:
                 pass
             try:
                 self.process.kill()
-            except:
+            except BaseException:
                 pass
             try:
                 self.process.wait()
-            except:
+            except BaseException:
                 pass
             self.process = None
+            try:
+                self.pout.close()
+            except BaseException:
+                pass
+            try:
+                self.pin.close()
+            except BaseException:
+                pass
             super(CpoSolverLocal, self).end()
 
 
+    def _is_abort_search_supported(self):
+        """ Check if this agent supports an actual abort_search() instead of killing the solver
+
+        Return:
+            True if this agent supports actual abort_search()
+        """
+        return self.proxy_version >= 17
+
+
     def _wait_event(self, xevt):
         """ Wait for a particular event while forwarding logs if any.
         Args:
@@ -415,6 +456,8 @@ class CpoSolverLocal(CpoSolverAgent):
         Raises:
             SolverException if an error occurs
         """
+        self.context.log(5, "Wait event: ", xevt)
+
         # Initialize first error string to enrich exception if any
         firsterror = None
 
@@ -426,6 +469,42 @@ class CpoSolverLocal(CpoSolverAgent):
             if evt == xevt:
                 return data
 
+            elif evt == EVT_CALLBACK_EVENT:
+                # First read the callback event
+                event = data
+                # Read event data in next message
+                evt, data = self._read_message()
+                assert evt == EVT_CALLBACK_DATA
+                res = self._create_result_object(CpoSolveResult, data)
+                self.solver._notify_callback_event(event, res)
+
+            elif evt == EVT_BLACKBOX_FUNCALL:
+                # Retrieve dialog id
+                sx = data.find(':')
+                if sx < 0:
+                    raise CpoException("Dialog id not found in blackbox function evaluation request '{}'".format(data))
+                did = data[:sx]
+                data = data[sx + 1:]
+                # Retrieve evaluation elements
+                try:
+                    bbf, args, bnds = self.solver._get_blackbox_function_eval_context(data)
+                    # Process blackbox function evaluation request
+                    if bbf.is_parallel_eval():
+                        # Evaluate in a separate thread
+                        thrd = threading.Thread(target=lambda: self._eval_blackbox(did, bbf, args, bnds))
+                        thrd.start()
+                    else:
+                        # Evaluate synchronously (ensure mutual exclusion)
+                        self._eval_blackbox(did, bbf, args, bnds)
+                except Exception as e:
+                    if self.context.log_exceptions:
+                        traceback.print_exc()
+                    # Build and send error message
+                    orig = traceback.extract_tb(sys.exc_info()[2])[-1]
+                    rdata = "{} ERROR ({}, {}) {}: {}".format(did, orig[0], orig[1], type(e).__name__, e)
+                    # Send back response
+                    self._write_message(EVT_BLACKBOX_RESULT, rdata)
+
             elif evt in (EVT_SOLVER_OUT_STREAM, EVT_SOLVER_WARN_STREAM):
                 if data:
                     # Store log if required
@@ -441,7 +520,7 @@ class CpoSolverLocal(CpoSolverAgent):
                     out.flush()
 
             elif evt == EVT_TRACE:
-                self.context.log(4, "ANGEL TRACE: " + data)
+                self.context.log(None, "ANGEL TRACE: " + data)
 
             elif evt == EVT_ERROR:
                 if firsterror is not None:
@@ -449,19 +528,38 @@ class CpoSolverLocal(CpoSolverAgent):
                 self.end()
                 raise CpoSolverException("Solver error: " + data)
 
-            elif evt == EVT_CALLBACK_EVENT:
-                event = data
-                # Read data
-                evt, data = self._read_message()
-                assert evt == EVT_CALLBACK_DATA
-                res = self._create_result_object(CpoSolveResult, data)
-                self.solver._notify_callback_event(event, res)
-
             else:
                 self.end()
                 raise CpoSolverException("Unknown event received from local solver: " + str(evt))
 
 
+    def _eval_blackbox(self, did, bbf, args, bnds):
+        """ Evaluate a blackbox function
+        Args:
+            did:  Dialog id
+            bbf:  Blackbox function descriptor
+            args: Argument values
+            bnds: Result bounds
+        """
+        # Process blackbox function evaluation request
+        try:
+            res = bbf._eval_function(bnds, *args)
+            # Build response
+            if res:
+                rdata = did + ' ' + (str(len(res)) if res else "0") + ' ' + (' '.join(str(v) for v in res))
+            else:
+                rdata = did + ' 0'
+        except Exception as e:
+            if self.context.log_exceptions:
+                traceback.print_exc()
+            # Build error message
+            orig = traceback.extract_tb(sys.exc_info()[2])[-1]
+            rdata = "{} ERROR ({}, {}) {}: {}".format(did, orig[0], orig[1], type(e).__name__, e)
+        # Send back response
+        #print("Response data: {}".format(rdata))
+        self._write_message(EVT_BLACKBOX_RESULT, rdata)
+
+
     def _wait_json_result(self, evt):
         """ Wait for a JSON result while forwarding logs if any.
         Args:
@@ -475,7 +573,7 @@ class CpoSolverLocal(CpoSolverAgent):
 
         # Store last json result
         self._set_last_json_result_string(data)
-        self.context.log(3, "JSON result:\n", data)
+        #self.context.log(3, "JSON result:\n", data)
 
         return self.last_json_result
 
@@ -535,7 +633,7 @@ class CpoSolverLocal(CpoSolverAgent):
             erline = frame + self._read_error_message()
             erline = erline.decode()
             self.end()
-            raise CpoSolverException("Invalid message header. Possible error generated by solver: " + erline)
+            raise CpoSolverException("Invalid message header. Possible error written by solver directly on stdout: " + erline)
 
         # Read message data
         tsize = decode_integer_big_endian_4(frame, 2)
@@ -562,7 +660,7 @@ class CpoSolverLocal(CpoSolverAgent):
         self.process_infos.incr(CpoProcessInfos.TOTAL_DATA_RECEIVE_SIZE, tsize + 6)
 
         # Log received message
-        self.context.log(5, "Read message: ", evt, ", data: '", data, "'")
+        self._log_received_message(evt, data)
 
         return evt, data
 
@@ -597,6 +695,8 @@ class CpoSolverLocal(CpoSolverAgent):
         # Return
         if IS_PYTHON_2:
             data = bytearray(data)
+        # print("Read frame: {}".format(', '.join([str(x) for x in data])))
+
         return data
 
 
@@ -625,7 +725,7 @@ class CpoSolverLocal(CpoSolverAgent):
         """ Send the model to the solver.
 
         Args:
-            copstr:  String containing the model in CPO format
+            cpostr:  String containing the model in CPO format
         """
         self._write_message(CMD_SET_CPO_MODEL, cpostr)
         self._wait_json_result(EVT_SUCCESS)  # JSON stored
@@ -642,41 +742,59 @@ class CpoSolverLocal(CpoSolverAgent):
         self._wait_event(EVT_SUCCESS)
 
 
+    def _register_blackbox_function(self, name, bbf):
+        """ Register a blackbox function in the solver
 
-###############################################################################
-##  Public functions
-###############################################################################
+        Args:
+            bbf: Blackbox function descriptor, object of class :class:`~docplex.cp.blackbox.CpoBlackboxFunction`
+        """
+        # Check angel version
+        aver = self.version_info.get('AngelVersion', 0)
+        if aver < 14:
+            raise CpoSolverException("This version of the CPO solver angel ({}) does not support blackbox functions.".format(aver))
 
-from docplex.cp.model import CpoModel
+        # Get blackbox function parameters
+        name = name.encode('utf8')
+        dimension = bbf.get_dimension()
+        argtypes = bbf.get_arg_types()
+        nbargs = len(argtypes)
 
-def get_solver_info():
-    """ Get the information data of the local CP solver that is target by the solver configuration.
-
-    This method creates a CP solver to retrieve this information, and end it immediately.
-    It returns a dictionary with various information, as in the following example:
-    ::
-    {
-       "AngelVersion" : 5,
-       "SourceDate" : "Sep 12 2017",
-       "SolverVersion" : "12.8.0.0",
-       "IntMin" : -9007199254740991,
-       "IntMax" : 9007199254740991,
-       "IntervalMin" : -4503599627370494,
-       "IntervalMax" : 4503599627370494,
-       "AvailableCommands" : ["Exit", "SetCpoModel", "SolveModel", "StartSearch", "SearchNext", "EndSearch", "RefineConflict", "Propagate", "RunSeeds"]
-    }
-
-    Returns:
-        Solver information dictionary, or None if not available.
-    """
-    try:
-        with CpoSolver(CpoModel()) as slvr:
-            if isinstance(slvr.agent, CpoSolverLocal):
-                return slvr.agent.version_info
-    except:
-        pass
-    return None
+        # Build message frame
+        frm = bytearray(2 + len(argtypes) + 5 + len(name) + 1)
 
+        # Add dimension and parameters
+        frm[0] = dimension
+        frm[1] = nbargs
+        for i in range(nbargs):
+            frm[2 + i] = BLACKBOX_ARGUMENT_TYPES_ENCODING[argtypes[i]]
 
+        # Add cache attributes
+        encode_integer_big_endian_4(bbf.cachesize, frm, 2 + nbargs)
+        frm[6 + nbargs] = 1 if bbf.globalcache else 0
+
+        # Add name
+        for i in range(len(name)):
+            frm[7 + nbargs + i] = name[i]
+        frm[7 + nbargs + len(name)] = 0
+
+        # Send message
+        self._write_message(CMD_ADD_BLACKBOX_FUN, data=frm)
+        self._wait_event(EVT_SUCCESS)
 
 
+    def _notify_execfile_not_found(self):
+        """ Print an error message if solver is not found """
+        out = self.context.get_log_output()
+        if out is None:
+            out = sys.stdout
+        banner = "#" * 79
+        out.write(banner + '\n')
+        out.write("# Solver executable file is not found !\n")
+        out.write("# Please check that:\n")
+        out.write("#  - you have installed IBM ILOG CPLEX Optimization Studio on your computer,\n")
+        out.write("#    (see https://ibmdecisionoptimization.github.io/docplex-doc/getting_started.html for details),\n")
+        out.write("#  - your system path includes a reference to the directory where the executable file 'cpoptimizer(.exe)' is located,\n")
+        out.write("#  - the context attribute 'context.solver.local.execfile' is properly set to 'cpoptimizer(.exe)',\n")
+        out.write("#  - or that it is set to an absolute path to this file.\n")
+        out.write(banner + '\n')
+        out.flush()
