@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
+# (c) Copyright IBM Corp. 2015 - 2022
 # --------------------------------------------------------------------------
 # Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis
 
@@ -25,12 +25,13 @@ import platform
 import copy
 import gzip
 import zipfile
+import traceback
 import importlib
 from collections import deque
 
 try:
     from collections.abc import Iterable  # Python >= 3.7
-except:
+except ImportError:
     from collections import Iterable      # Python < 3.7
 
 try:
@@ -39,12 +40,16 @@ except ImportError:
     from io import StringIO        # Python 3
 
 
-###############################################################################
-## Constants
-###############################################################################
+#-----------------------------------------------------------------------------
+#  Constants
+#-----------------------------------------------------------------------------
+
+# Python main version number
+PYTHON_MAIN_VERSION = sys.version_info[0]
 
 # Python 2 indicator
-IS_PYTHON_2 = (sys.version_info[0] == 2)
+IS_PYTHON_2 = (PYTHON_MAIN_VERSION == 2)
+
 
 # Check if numpy is available
 try:
@@ -71,15 +76,15 @@ DEFAULT = "default"
 IS_WINDOWS = platform.system() == 'Windows'
 
 
-###############################################################################
-## Public classes
-###############################################################################
+#-----------------------------------------------------------------------------
+#  Public classes
+#-----------------------------------------------------------------------------
 
 class CpoException(Exception):
     """ Exception thrown in case of CPO errors.
     """
     def __init__(self, msg):
-        """ Create a new exception
+        """ **Constructor**
 
         Args:
             msg: Error message
@@ -91,7 +96,7 @@ class CpoNotSupportedException(CpoException):
     """ Exception thrown when a CPO function is not supported.
     """
     def __init__(self, msg):
-        """ Create a new exception
+        """ **Constructor**
 
         Args:
             msg: Error message
@@ -102,7 +107,7 @@ class CpoNotSupportedException(CpoException):
 class Context(dict):
     """ Class handling miscellaneous list of parameters. """
     def __init__(self, **kwargs):
-        """ Create a new context
+        """ **Constructor**
 
         Args:
             **kwargs: List of key=value to initialize context with.
@@ -161,12 +166,11 @@ class Context(dict):
         if name.startswith('__'):
             raise AttributeError
         ctx = self
-        while True:
+        while ctx is not None:
             if name in ctx:
                return ctx.get(name)
             ctx = ctx.get_parent()
-            if ctx is None:
-                return default
+        return default
 
 
     def del_attribute(self, name):
@@ -308,8 +312,8 @@ class Context(dict):
         return res
 
 
-    def add(self, ctx):
-        """ Add another context to this one.
+    def set_other(self, ctx):
+        """ Set another context into this one.
 
         All attributes of given context are set in this one, replacing previous value if any.
         If one value is another context, it is cloned before being set.
@@ -323,6 +327,11 @@ class Context(dict):
             self.set_attribute(k, v)
 
 
+    def add(self, ctx):
+        # Obsolete implementation, kept to assure ascending compatibility
+        self.set_other(ctx)
+
+
     def is_log_enabled(self, vrb):
         """ Check if log is enabled for a given verbosity
 
@@ -363,7 +372,8 @@ class Context(dict):
             prfx = self.log_prefix
             if prfx:
                 out.write(str(prfx))
-            out.write(''.join([str(m) for m in msg]) + "\n")
+            out.write(''.join([str(m) for m in msg]))
+            out.write("\n")
             out.flush()
 
 
@@ -424,6 +434,126 @@ class Context(dict):
         out.flush()
 
 
+    def export_flat(self, out=None):
+        """ Export this context in flat format
+
+        Each context attribute is written on a single line <path>=<value> with UTF-8 encoding.
+
+        Args:
+            out (Optional):  Target output stream or file name. If not given, default value is sys.stdout.
+        """
+        # Check file
+        if is_string(out):
+            with open_utf8(os.path.abspath(out), mode='w') as f:
+                self.export_flat(f)
+                return
+        # Check default output
+        if out is None:
+            out = sys.stdout
+
+        # Build dictionary of all attributes by pathes
+        adict = {}
+        self._build_path_dict('', adict)
+
+        # Print values
+        for k in sorted(adict.keys()):
+            out.write(make_unicode(k))
+            out.write(u" = ")
+            v = adict[k]
+            if is_string(v):
+                out.write(to_printable_string(v))
+            else:
+                out.write(make_unicode(str(v)))
+            out.write(u'\n')
+
+
+    def export_flat_as_string(self):
+        """ Export this context in flat format as a string.
+
+        Each context attribute is written on a single line <path>=<value>
+
+        Args:
+            out (Optional):  Target output stream or file name. If not given, default value is sys.stdout.
+        """
+        out = StringIO()
+        self.export_flat(out)
+        res = out.getvalue()
+        out.close()
+        return res
+
+
+    def import_flat(self, inp=None):
+        """ Import a flat file in this context
+
+        Each context attribute is added in this context.
+        Each value is converted in the most appropriate type: string, integer, float, boolean or None.
+        Only atomic values are allowed, not lists or other Python complex objects.
+
+        Args:
+            inp:  Input stream or file
+        """
+        # Check file
+        if is_string(inp):
+            with open_utf8(os.path.abspath(inp), mode='r') as f:
+                self.import_flat(f)
+                return
+
+        # Read all lines
+        for line in inp.readlines():
+            # Check empty and comment lines
+            line = line.strip()
+            if not line or line.startswith('#'):
+                continue
+            # Get attribute path
+            cx = line.find('=')
+            if cx < 0:
+                raise Exception("No equal sign found in line '{}'".format(line))
+            k = line[:cx].strip()
+            v = line[cx+1:].strip()
+            v = string_to_value(v)
+            self.set_by_path(k, v)
+
+
+    def _build_path_dict(self, path, result):
+        """ Build dictionary of all context values with pathes as keys
+
+        Args:
+            path:  Path of this context
+            result: Result dictionary to fill
+        """
+        # Build path prefix
+        if path:
+            path = path + "."
+        for k in sorted(self.keys()):
+            v = self[k]
+            if isinstance(v, Context):
+                v._build_path_dict(path + k, result)
+            else:
+                result[path + k] = v
+
+
+class PersistentContext(Context):
+    """ Persistent context stored in a file. """
+
+    def __init__(self, cfile):
+        """ **Constructor**
+
+        If the given file exists, it is read in this context.
+
+        Args:
+            cfile: File where context is stored with its flat representation
+        """
+        super(PersistentContext, self).__init__()
+        vars(self)['storage_file'] = cfile
+        if os.path.isfile(cfile):
+            self.import_flat(cfile)
+
+    def save(self):
+        """ Save this context in the file.
+        """
+        self.export_flat(vars(self)['storage_file'])
+
+
 class IdAllocator(object):
     """ Allocator of identifiers.
 
@@ -437,7 +567,7 @@ class IdAllocator(object):
     DIGITS = "0123456789"
     LETTERS_AND_DIGITS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
     def __init__(self, prefix, bdgts="0123456789"):
-        """ Create a new id allocator.
+        """ **Constructor**
 
         Args:
             prefix:  Prefix of all ids
@@ -494,7 +624,7 @@ class SafeIdAllocator(object):
                  'lock',    # Lock to protect counter
                  )
     def __init__(self, prefix, bdgts="0123456789"):
-        """ Create a new id allocator.
+        """ **Constructor**
 
         Args:
             prefix:  Prefix of all ids
@@ -547,7 +677,7 @@ class SafeIdAllocator(object):
 
 
 class KeyIdDict(object):
-    """ Dictionary using id of the key objects as key.
+    """ Dictionary using internally id of the key objects as dictionary key.
 
     This object allows to use any Python object as key (with no __hash__() function),
     and to map a value on the physical instance of the key.
@@ -556,7 +686,6 @@ class KeyIdDict(object):
                  )
 
     def __init__(self):
-        super(KeyIdDict, self).__init__()
         self.kdict = {}
 
     def set(self, key, value):
@@ -613,7 +742,6 @@ class ObjectCache(object):
                  )
 
     def __init__(self, maxsize):
-        super(ObjectCache, self).__init__()
         self.obj_dict = {}
         self.max_size = maxsize
         self.key_list = deque()
@@ -807,7 +935,8 @@ class Chrono(object):
     __slots__ = ('startTime',  # Chrono start time
                  )
     def __init__(self):
-        """ Create a new chronometer initialized with current time.
+        """ **Constructor**
+         Create a new chronometer initialized with current time.
         """
         super(Chrono, self).__init__()
         self.restart()
@@ -854,7 +983,7 @@ class Barrier(object):
                  'barrier'   # Threads blocking lock
                  )
     def __init__(self, parties):
-        """ Create a new barrier
+        """ **Constructor**
 
         Args:
             parties:  Number of parties required before unlocking the barrier
@@ -888,7 +1017,8 @@ class FunctionCache(object):
                  'funct',   # Function whose results is cached.
                  )
     def __init__(self, f):
-        """
+        """ **Constructor**
+
         Args:
             f:  Function whose results should be cached.
         """
@@ -1022,9 +1152,9 @@ class ListDict(dict):
 
 
 
-###############################################################################
-## Public functions
-###############################################################################
+#-----------------------------------------------------------------------------
+#  Public functions
+#-----------------------------------------------------------------------------
 
 def check_default(val, default):
     """ Check that an argument value is DEFAULT and returns the default value if so.
@@ -1178,6 +1308,72 @@ def to_string(val):
     return str(val)
 
 
+_DEFAULT_STRING_VALUES = {"none": None, "true": True, "false": False}
+def string_to_value(s):
+    """ Convert a string in its most representative python value
+
+    If the string is encapsulated in double_quotes, the internal string is returned.
+    If string in lower case is "none", None is returned.
+    If string in lower case is "true" or "false", True or False is returned.
+    If string matches an integer, the integer is returned.
+    If the string matches a float, the float is returned
+    Otherwise, string is returned as it is
+
+    Args:
+        s: String to convert
+    Returns:
+        Value corresponding to the string
+    """
+    # Check first character
+    c = s[0]
+    ls = len(s)
+    if c == '"' and ls > 1 and s[-1] == '"':
+        return s[1:ls-1]
+    if c.isdigit() or (c == '-' and ls >= 2 and s[1].isdigit()):
+        # Try to convert as number
+        try:
+            # Check if possible float
+            if s.find('.') >= 0:
+                s = float(s)
+            else:
+                s = int(s)
+        except:
+            pass
+    else:
+        ns = s.lower()
+        if ns in _DEFAULT_STRING_VALUES:
+            return _DEFAULT_STRING_VALUES[ns]
+    return s
+
+def to_compact_SI(val):
+    """ Convert a long value in a string using suffix k, M or G to display only 3 digits.
+
+    Args:
+        val:  Integer value to convert
+    Returns:
+        String representing this value with 3 significant digits
+    """
+    if val < 1000:
+        return str(val)
+    if val < 10000:
+        return str(val // 1000) + "." + str((val % 1000) // 10) + "k"
+    if val < 100000:
+        return str(val // 1000) + "." + str((val % 1000) // 100) + "k"
+    if val < 1000000:
+        return str(val // 1000) + "k"
+    if val < 10000000:
+        return str(val // 1000000) + "." + str((val % 1000000) // 10000) + "M"
+    if val < 100000000:
+        return str(val // 1000000) + "." + str((val % 1000000) // 100000) + "M"
+    if val < 1000000000:
+        return str(val // 1000000) + "M"
+    if val < 10000000000:
+        return str(val // 1000000000) + "." + str((val % 1000000000) // 10000000) + "G"
+    if val < 100000000000:
+        return str(val // 1000000000) + "." + str((val % 1000000000) // 100000000) + "G"
+    return str(val // 1000000000) + "G"
+
+
 def _get_vars(obj):
     """ Get the list variable names of an object.
     """
@@ -1402,7 +1598,7 @@ def list_module_public_functions(mod, excepted=()):
     """ Build the list of all public functions of a module.
 
     Args:
-        mod:  Module to parse
+        mod:       Module to parse
         excepted:  List of function names to not include. Default is none.
     Returns:
         List of public functions declared in this module
@@ -1445,8 +1641,9 @@ if IS_PYTHON_2:
 # Add numpy types if any
 if IS_NUMPY_AVAILABLE:
     BOOL_TYPES.add(numpy.bool_)
-    BOOL_TYPES.add(numpy.bool)
-
+    if numpy.__version__ < '1.20':
+        BOOL_TYPES.add(numpy.bool)
+        
     INTEGER_TYPES.add(numpy.int_)
     INTEGER_TYPES.add(numpy.intc)
     INTEGER_TYPES.add(numpy.intp)
@@ -1577,6 +1774,17 @@ else:
         return isinstance(val, (list, tuple))
 
 
+def is_tuple(val):
+    """ Check if a value is a tuple.
+
+    Args:
+        val: Value to check
+    Returns:
+        True if value is a tuple
+    """
+    return isinstance(val, tuple)
+
+
 if IS_PANDA_AVAILABLE:
 
     def is_panda_series(val):
@@ -1704,22 +1912,51 @@ def is_symbol_char(c):
     return c in _SYMBOL_CHARS
 
 
-def to_printable_string(id):
-    """ Build a printable string from raw string (add escape sequences and quotes if necessary).
+def is_symbol(s):
+    """ Check whether a string can be used in a symbol.
 
     Args:
-        id: Identifier string
+        s: String to check
+    Returns:
+        True if character string can be used as s symbol
+    """
+    return s and is_string(s) and (not s[0] in _DIGIT_CHARS) and all((c in _SYMBOL_CHARS) for c in s)
+
+
+def to_printable_id(s):
+    """ Build a CPO printable identifier from its raw string (add escape sequences and quotes if necessary).
+
+    Args:
+        s: Identifier string
     Returns:
         Unicode CPO identifier string, including double quotes and escape sequences if needed if not only chars and integers
     """
     # Check empty string
-    if len(id) == 0:
+    if len(s) == 0:
         return u'""'
     # Check if string can be used as it is
-    if (all((c in _SYMBOL_CHARS) for c in id)) and not id[0] in _DIGIT_CHARS:
-        return make_unicode(id)
+    if (not s[0] in _DIGIT_CHARS) and all((c in _SYMBOL_CHARS) for c in s):
+        return make_unicode(s)
     # Build result string
-    return u'"' + ''.join(_TO_SPECIAL_CHARS.get(c, c) for c in id) + u'"'
+    return u'"' + ''.join(_TO_SPECIAL_CHARS.get(c, c) for c in s) + u'"'
+
+
+def to_printable_string(s):
+    """ Build a printable string from raw string (add escape sequences and quotes if necessary).
+
+    Args:
+        str: String to convert
+    Returns:
+        Unicode string, including always double quotes and escape sequences if needed
+    """
+    # Check empty string
+    if len(s) == 0:
+        return u'""'
+    # Check if string can be used as it is
+    if (all((c in _SYMBOL_CHARS) for c in s)) and not s[0] in _DIGIT_CHARS:
+        return u'"' + make_unicode(s) + u'"'
+    # Build result string
+    return u'"' + ''.join(_TO_SPECIAL_CHARS.get(c, c) for c in s) + u'"'
 
 
 def to_internal_string(strg):
@@ -1777,7 +2014,7 @@ if IS_PYTHON_2:
         Returns:
             String in unicode
         """
-        return s if type(s) is unicode else unicode(s)
+        return unicode(s)
 else:
     def make_unicode(s):
         """ Convert a string in unicode.
@@ -1833,17 +2070,35 @@ def is_exe_file(f):
     return os.path.isfile(f) and os.access(f, os.X_OK)
 
 
+def get_main_file():
+    """ Get the name of the main Python module
+
+    Returns:
+        Absolute path of the Python module that has been used as main.
+        None if unknown.
+    """
+    try:
+        return os.path.abspath(sys.modules['__main__'].__file__)
+    except:
+        try:
+            return os.path.abspath(sys.argv[0])
+        except:
+            pass
+    return None
+
+
 def get_system_path():
-    """ Get the system path as a list of files
+    """ Get the system path as a list of directory names
 
     Returns:
         List of names in the system path
     """
-    path = os.getenv('PATH')
-    if path:
-        return path.split(os.pathsep)
-    return []
-
+    rpath = []
+    for pn in ('LD_LIBRARY_PATH', 'DYLD_LIBRARY_PATH', 'PATH'):
+        path = os.getenv(pn)
+        if path:
+            rpath.extend(path.split(os.pathsep))
+    return rpath
 
 
 def search_exec_file(f, path):
@@ -1966,6 +2221,33 @@ def get_module_version(mname):
         return None
 
 
+def is_in_notebook():
+    """ Check whether program is currently running inside a IPython or Jupyter notebook.
+
+    Returns:
+        True if the program is running in a notebook, false otherwise.
+    """
+    rstk = traceback.extract_stack(limit=1)[0]
+    return rstk[0].startswith("<ipython")
+
+
+def write_checking_unicode_errors(out, data):
+    """ Write data on a stream, and by-pass UnicodeEncodeError if any.
+
+    Args:
+        out:  Output stream
+        data: Data to write
+    """
+    try:
+        out.write(data)
+    except UnicodeEncodeError:
+        # Determine proper encoding
+        encoding = out.encoding
+        if not encoding:
+            encoding = 'ascii'
+        # Rewrite data encoding it explicitly
+        out.write(data.encode(encoding, errors='backslashreplace').decode(encoding, errors='backslashreplace'))
+
 #-----------------------------------------------------------------------------
 # Zip iterator functions to scan lists simultaneously
 #-----------------------------------------------------------------------------
