@@ -1,56 +1,36 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2016
+# (c) Copyright IBM Corp. 2022
 # --------------------------------------------------------------------------
 
-from six import iteritems
+# gendoc: ignore
 
 from docplex.mp.model import Model
 from docplex.mp.aggregator import ModelAggregator
 from docplex.mp.quad import VarPair
 from docplex.mp.utils import is_number, is_iterable, generate_constant, \
-    is_pandas_dataframe, is_pandas_series, is_numpy_matrix, is_scipy_sparse, is_ordered_sequence
+    is_pandas_dataframe, is_numpy_matrix,\
+    ordered_sequence_to_list
+from docplex.mp.ds_utils import is_scipy_sparse
 from docplex.mp.constants import ComparisonType
 
-from docplex.mp.compat23 import izip
-from docplex.mp.error_handler import docplex_fatal
 from docplex.mp.xcounter import update_dict_from_item_value
 
+
 class AdvAggregator(ModelAggregator):
-    def __init__(self, linear_factory, quad_factory):
-        ModelAggregator.__init__(self, linear_factory, quad_factory)
-
-    def _scal_prod_vars_all_different(self, terms, coefs):
-        checker = self._checker
-        if not is_iterable(coefs, accept_string=False):
-            checker.typecheck_num(coefs)
-            return coefs * self._sum_vars_all_different(terms)
-        else:
-            # coefs is iterable
-            lcc_type = self.counter_type
-            lcc = lcc_type()
-            lcc_setitem = lcc_type.__setitem__
-            number_validation_fn = checker.get_number_validation_fn()
-            if number_validation_fn:
-                for dvar, coef in izip(terms, coefs):
-                    safe_coef = number_validation_fn(coef)
-                    if safe_coef:
-                        lcc_setitem(lcc, dvar, safe_coef)
-            else:
-                for dvar, coef in izip(terms, coefs):
-                    if coef:  # zero test is much cheaper than a setitem
-                        lcc_setitem(lcc, dvar, coef)
-
-            return self._to_expr(qcc=None, lcc=lcc)
 
     def scal_prod_triple(self, left_terms, right_terms, coefs):
         used_coefs = None
         checker = self._model._checker
+        single_coeff = False
+        single_left = False
+        single_right = False
 
         if is_iterable(coefs, accept_string=False):
             used_coefs = coefs
         elif is_number(coefs):
+            single_coeff = True
             if coefs:
                 used_coefs = generate_constant(coefs, count_max=None)
             else:
@@ -59,18 +39,26 @@ class AdvAggregator(ModelAggregator):
             self._model.fatal("scal_prod_triple expects iterable or number as coefficients, got: {0!r}", coefs)
 
         if is_iterable(left_terms):
-            used_left = checker.typecheck_var_seq(left_terms)
+            l_terms = list(left_terms)
+            for l, lterm in enumerate(l_terms):
+                checker.typecheck_operand(lterm, accept_numbers=False, caller="AdvModel.scal_prod_triple.left_terms")
+            used_left = l_terms
         else:
+            single_left = True
             checker.typecheck_var(left_terms)
             used_left = generate_constant(left_terms, count_max=None)
 
         if is_iterable(right_terms):
-            used_right = checker.typecheck_var_seq(right_terms)
+            l_right = list(right_terms)
+            for r, rterm in enumerate(l_right):
+                checker.typecheck_operand(rterm, accept_numbers=False, caller="AdvModel.scal_prod_triple.right_terms")
+            used_right = l_right
         else:
+            single_right = True
             checker.typecheck_var(right_terms)
             used_right = generate_constant(right_terms, count_max=None)
 
-        if used_coefs is not coefs and used_left is not left_terms and used_right is not right_terms:
+        if all((single_coeff, single_left, single_right)):
             # LOOK
             return left_terms * right_terms * coefs
 
@@ -79,10 +67,10 @@ class AdvAggregator(ModelAggregator):
     def _scal_prod_triple(self, coefs, left_terms, right_terms):
         # INTERNAL
         accumulated_ct = 0
-        qcc = self.counter_type()
-        lcc = self.counter_type()
+        qcc = self._term_dict_type()
+        lcc = self._term_dict_type()
         number_validation_fn = self._checker.get_number_validation_fn()
-        for coef, lterm, rterm in izip(coefs, left_terms, right_terms):
+        for coef, lterm, rterm in zip(coefs, left_terms, right_terms):
             if coef:
                 safe_coef = number_validation_fn(coef) if number_validation_fn else coef
                 lcst = lterm.get_constant()
@@ -104,15 +92,15 @@ class AdvAggregator(ModelAggregator):
     def _scal_prod_triple_vars(self, coefs, left_terms, right_terms):
         # INTERNAL
         # assuming all arguments are iterable.
-        dcc = self.counter_type
+        dcc = self._term_dict_type
         qcc = dcc()
         number_validation_fn = self._checker.get_number_validation_fn()
         if number_validation_fn:
-            for coef, lterm, rterm in izip(coefs, left_terms, right_terms):
+            for coef, lterm, rterm in zip(coefs, left_terms, right_terms):
                 safe_coef = number_validation_fn(coef) if number_validation_fn else coef
                 update_dict_from_item_value(qcc, VarPair(lterm, rterm), safe_coef)
         else:
-            for coef, lterm, rterm in izip(coefs, left_terms, right_terms):
+            for coef, lterm, rterm in zip(coefs, left_terms, right_terms):
                 update_dict_from_item_value(qcc, VarPair(lterm, rterm), coef)
         return self._to_expr(qcc=qcc)
 
@@ -124,12 +112,6 @@ class AdvAggregator(ModelAggregator):
             qcc_setitem(qcc, VarPair(t), 1)
         return self._to_expr(qcc=qcc)
 
-    def _sumsq_vars(self, dvars):
-        qcc = self._quad_factory.term_dict_type()
-        for v in dvars:
-            update_dict_from_item_value(qcc, VarPair(v), 1)
-        return self._to_expr(qcc=qcc)
-
     def quad_matrix_sum(self, matrix, lvars, symmetric=False):
         # assume matrix is a NxN matrix
         # vars is a N-vector of variables
@@ -173,12 +155,6 @@ class AdvAggregator(ModelAggregator):
 
         return self._to_expr(qcc=qterms)
 
-    def vector_compare(self, left_exprs, right_exprs, sense):
-        lfactory = self._linear_factory
-        assert len(left_exprs) == len(right_exprs)
-        cts = [lfactory._new_binary_constraint(left, sense, right) for left, right in izip(left_exprs, right_exprs)]
-        return cts
-
 
 # noinspection PyProtectedMember
 class AdvModel(Model):
@@ -186,10 +162,10 @@ class AdvModel(Model):
     This class is a specialized version of the :class:`docplex.mp.model.Model` class with useful non-standard modeling
     functions.
     """
-    _fast_settings = {'keep_ordering': False, 'checker': 'off', 'keep_all_exprs': False}
+    _fast_settings = {'keep_ordering': False, 'keep_all_exprs': False}
 
     def __init__(self, name=None, context=None, **kwargs):
-        for k, v in iteritems(self._fast_settings):
+        for k, v in self._fast_settings.items():
             if k not in kwargs:
                 # force fast settings if not present
                 kwargs[k] = v
@@ -197,31 +173,18 @@ class AdvModel(Model):
         Model.__init__(self, name=name, context=context, **kwargs)
         self._aggregator = AdvAggregator(self._lfactory, self._qfactory)
 
-    def _prepare_constraint(self, ct, ctname, check_for_trivial_ct, arg_checker=None):
-        # INTERNAL
-        if ct is False:
-            # happens with sum([]) and constant e.g. sum([]) == 2
-            msg = "Adding a trivially infeasible constraint"
-            if ctname:
-                msg += ' with name: {0}'.format(ctname)
-            # analogous to 0 == 1, model is sure to fail
-            self.fatal(msg)
+    def _check_trivial_constraints(self):
+        # disable all checks of trivial constraints.
+        return False
 
-        if ct is True:
-            return False
+    def _typecheck_in_model(self, ct, caller):
+        pass
 
-
-        # --- name management ---
-        if ctname:
-            ct_name_map = self._cts_by_name
-            if ct_name_map is not None:
-                ct_name_map[ctname] = ct
-            ct.name = ctname
-        # ---
+    def _check_new_ct_index(self, ct):
         return True
 
     def sumsq_vars(self, terms):
-        return self._aggregator._sumsq_vars(terms)
+        return self._aggregator._sumsq(terms)
 
     def sumsq_vars_all_different(self, terms):
         """
@@ -265,14 +228,10 @@ class AdvModel(Model):
 
         Note:
             To improve performance, the check for duplicates can be turned off by setting
-            `checker='none'` in the `kwargs` of the :class:`docplex.mp.model.Model` object. As this argument
+            `checker='off'` in the `kwargs` of the :class:`docplex.mp.model.Model` object. As this argument
             turns off checking everywhere, it should be used with extreme caution.
         """
-        self._checker.check_ordered_sequence(arg=terms,
-                                             caller='Model.scal_prod() requires a list of expressions/variables')
-        var_seq = self._checker.typecheck_var_seq_all_different(terms)
-        return self._aggregator._scal_prod_vars_all_different(var_seq, coefs)
-
+        return super(AdvModel, self).scal_prod_vars_all_different(terms, coefs)
 
     def quad_matrix_sum(self, matrix, dvars, symmetric=False):
         """
@@ -302,7 +261,7 @@ class AdvModel(Model):
         Example:
             `Model.quad_matrix_sum([[[1, 2], [3, 4]], [x, y])` returns the expression `x^2+4y^2+5x*yt`.
         """
-        lvars = AdvModel._to_list(dvars, caller='Model.quad_matrix_sum')
+        lvars = ordered_sequence_to_list(dvars, caller='Model.quad_matrix_sum')
         self._checker.typecheck_var_seq(lvars)
         if is_scipy_sparse(matrix):
             return self._aggregator._sparse_quad_matrix_sum(matrix, lvars, symmetric=symmetric)
@@ -346,7 +305,7 @@ class AdvModel(Model):
         This method is faster than the standard generic scalar quadratic product method due to the fact that it takes only variables and does not take expressions as arguments.
 
         Example:
-            `Model.scal_prod_vars_triple([x, y], [z, t], [2, 3])` returns the expression `2xz + 3yt`.
+            `Model.scal_prod_triple_vars([x, y], [z, t], [2, 3])` returns the expression `2xz + 3yt`.
 
         :param left_terms: A list or an iterator on variables.
         :param right_terms: A list or an iterator on variables.
@@ -392,16 +351,6 @@ class AdvModel(Model):
             return self._aggregator._scal_prod_triple_vars(left_terms=used_left,
                                                            right_terms=used_right, coefs=used_coefs)
 
-    @classmethod
-    def _to_list(cls, s, caller):
-        if is_pandas_series(s):
-            return s.tolist()
-        elif is_ordered_sequence(s):
-            return s
-        else:
-            docplex_fatal('{0} requires ordered sequences: lists, numpy array or Series, got: {1}', caller, type(s))
-            return list(s)
-
     def matrix_constraints(self, coef_mat, dvars, rhs, sense='le'):
         """
         Creates a list of linear constraints
@@ -417,8 +366,6 @@ class AdvModel(Model):
         <op> is the comparison operator that defines the sense of the constraint. By default, this generates
         a 'less-than-or-equal' constraint.
 
-        Example:
-            `Model.scal_prod_vars_triple([x, y], [z, t], [2, 3])` returns the expression `2xz + 3yt`.
 
         :param coef_mat: A matrix of coefficients with M rows and N columns. This argument accepts
             either a list of lists of numbers, a `numpy` array with size (M,N), or a `scipy` sparse matrix.
@@ -443,11 +390,11 @@ class AdvModel(Model):
 
             then::
 
-                `mdl.matrix_constraint(A, X, B, 'GE')` returns a list of two constraints
+                `mdl.matrix_constraints(A, X, B, 'ge')` returns a list of two constraints
                 [(x + 2y+3z <= 100), (4x + 5y +6z <= 200)].
 
         Note:
-            If the dimensions of the matrix and variables or of the matrix and number sequence do not match,
+            If the dimensions of the matrix and variables, or if the matrix and rhs shapes do not match,
             an error is raised.
 
         """
@@ -472,8 +419,8 @@ class AdvModel(Model):
             except AttributeError:
                 self.fatal('All columns should have a len()')
 
-        s_dvars = self._to_list(dvars, caller='Model.matrix-constraints()')
-        s_rhs = self._to_list(rhs, caller='Model.matrix-constraints()')
+        s_dvars = ordered_sequence_to_list(dvars, caller='Model.matrix-constraints()')
+        s_rhs = ordered_sequence_to_list(rhs, caller='Model.matrix-constraints()')
         # check
 
         checker.typecheck_var_seq(s_dvars)
@@ -532,7 +479,7 @@ class AdvModel(Model):
 
             then::
 
-                `mdl.range_constraints(A, X, L, U)` returns a list of two ranges
+                `mdl.matrix_ranges(A, X, L, U)` returns a list of two ranges
                 [(101 <= x + 2y+3z <= 102), (201 <= 4x + 5y +6z <= 202)].
 
         Note:
@@ -561,9 +508,9 @@ class AdvModel(Model):
             except AttributeError:
                 self.fatal('All columns should have a len()')
 
-        s_dvars = self._to_list(dvars, caller='Model.range_constraints()')
-        s_lbs = self._to_list(lbs, caller='Model.range_constraints()')
-        s_ubs = self._to_list(ubs, caller='Model.range_constraints()')
+        s_dvars = ordered_sequence_to_list(dvars, caller='Model.range_constraints()')
+        s_lbs = ordered_sequence_to_list(lbs, caller='Model.range_constraints()')
+        s_ubs = ordered_sequence_to_list(ubs, caller='Model.range_constraints()')
         # check
 
         checker.typecheck_var_seq(s_dvars)
@@ -588,21 +535,3 @@ class AdvModel(Model):
             return self._aggregator._sparse_matrix_ranges(coef_mat, s_dvars, s_lbs, s_ubs)
         else:
             return self._aggregator._matrix_ranges(coef_mat, s_dvars, s_lbs, s_ubs)
-
-    def vector_compare(self, lhss, rhss, sense):
-        l_lhs = self._to_list(lhss, caller='Model.vector.compare')
-        l_rhs = self._to_list(rhss, caller='Model.vector.compare')
-        if len(l_lhs) != len(l_rhs):
-            self.fatal('Model.vector_compare() got sequences with different length, left: {0}, right: {1}'.
-                       format(len(l_lhs), len(l_rhs)))
-        ctsense = ComparisonType.parse(sense)
-        return self._aggregator.vector_compare(l_lhs, l_rhs, ctsense)
-
-    def vector_compare_le(self, lhss, rhss):
-        return self.vector_compare(lhss, rhss, 'le')
-
-    def vector_compare_ge(self, lhss, rhss):
-        return self.vector_compare(lhss, rhss, 'ge')
-
-    def vector_compare_eq(self, lhss, rhss):
-        return self.vector_compare(lhss, rhss, 'eq')
