@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
 import operator
@@ -124,20 +124,22 @@ class RelaxationMode(Enum):
     searches for an optimal solution while keeping the relaxation at the minimal value found in phase 1.
 
     Enumerated values work in pairs: MinXXX, OptXXX
+    
         - MinXXX values stop at phase 1, they look for a feasible soluion, they do not
-            optimize the objective.
+          optimize the objective.
         - OptXXX run the two phases, looking for an optimal relaxed solution. They take longer.
 
     The metric used to evaluate the quality of the relaxation is determined by the XXX part of
     the name. There are three metrics:
-      - Inf (MinInf, OptInf) minimizes the number of relaxed constraints.  This metric
-        will prefer to relax one constraint, even with a huge slack, instead of two.
-      - Sum (MinSum, OptSum): minimizes the sum of relaxations.
-      - Quand (MinQuad, OptQuad): minimizes the sum of squares of relaxations.
-        This metric is the most expensive in computation time,
-        but avoids huge discrepancies between relaxations:
-        two constraints with relaxations of 2,2 will have a better quality (2^2 + 2^2 = 8)
-        than relaxations of 3,1 (3^2 +1 = 10).
+    
+        - Inf (MinInf, OptInf) minimizes the number of relaxed constraints.  This metric
+          will prefer to relax one constraint, even with a huge slack, instead of two.
+        - Sum (MinSum, OptSum): minimizes the sum of relaxations.
+        - Quand (MinQuad, OptQuad): minimizes the sum of squares of relaxations.
+          This metric is the most expensive in computation time,
+          but avoids huge discrepancies between relaxations:
+          two constraints with relaxations of 2,2 will have a better quality (2^2 + 2^2 = 8)
+          than relaxations of 3,1 (3^2 +1 = 10).
     """
 
     MinSum, OptSum, MinInf, OptInf, MinQuad, OptQuad = range(6)
@@ -301,10 +303,6 @@ class ObjectiveSense(Enum):
         """
         return self is ObjectiveSense.Maximize
 
-    @property
-    def cplex_coef(self):
-        return 1 if self.is_minimize() else -1
-
     @property
     def verb(self):
         """ Returns a string describing the objective (in lowercase)
@@ -317,11 +315,14 @@ class ObjectiveSense(Enum):
 
     @property
     def short_name(self):
-        """ Returns a short (three letters) string describing the objective: min or max
-
+        """ Returns a short ('min' or 'max') string describing the objective.
         """
         return self.verb[:3]
 
+    @property
+    def cplex_coef(self):
+        return 1 if self.is_minimize() else -1
+
     @classmethod
     def from_cplex(cls, cpx_sense):
         if cpx_sense == 1:
@@ -332,7 +333,19 @@ class ObjectiveSense(Enum):
             raise ValueError("expecting +1 or -1, {0} was passed".format(cpx_sense))
 
     @staticmethod
-    def parse(arg, logger=None, default_sense=None):
+    def parse(arg, logger=None):
+        """ Converts an argument to an objective sense.
+
+        Accepts either
+        - an objective sense (returns itself)
+        - a string (in  long or short format),
+        - an integer (in CPLEX convention, 1 is for minimize, -1 for maximize).
+
+        :param arg: the argument to convert to a sense
+        :param logger:
+
+        :return: An instance of enumerated `ObjectiveSense` class.
+        """
         if isinstance(arg, ObjectiveSense):
             return arg
 
@@ -342,28 +355,22 @@ class ObjectiveSense(Enum):
                 return ObjectiveSense.Minimize
             elif lower_text in {"maximize", "max"}:
                 return ObjectiveSense.Maximize
-            elif default_sense:
-                logger.error(
-                    "Text is not recognized as objective sense: {0}, expecting \"min\" or \"max\" - using default {1:s}",
-                    (arg, default_sense))
-                return default_sense
             elif logger:
-                logger.fatal("Text is not recognized as objective sense: {0}, expecting ""min"" or ""max", (arg,))
+                logger.fatal("Not an objective sense: '{0}', expecting 'min|max'", arg)
             else:
-                docplex_fatal("Text is not recognized as objective sense: {0}, expecting ""min"" or ""max".format(arg))
+                docplex_fatal("Not an objective sense: '{0}', expecting 'min|max'".format(arg))
         elif is_int(arg):
             if arg == 1:
                 return ObjectiveSense.Minimize
             elif -1 == arg:
                 return ObjectiveSense.Maximize
             else:
-                logger.fatal("cannot convert: <{}> to objective sense", (arg,))
-        elif arg is None:
-            return default_sense
-        elif logger:
-            logger.fatal("cannot convert: <{}> to objective sense", (arg,))
+                logger.fatal("Not an objective sense: <{}>", (arg,))
+
+        if logger:
+            logger.fatal("Not an objective sense: <{}>", (arg,))
         else:
-            docplex_fatal("cannot convert: <{}> to objective sense".format(arg))
+            docplex_fatal("Not an objective sense: <{}>".format(arg))
 
 
 # noinspection PyPep8
@@ -379,11 +386,14 @@ class CplexScope(Enum):
     # INTERNAL
     VAR_SCOPE = 0, 'x', 'variables'
     LINEAR_CT_SCOPE = 1, 'c', 'linear constraints'
-    IND_CT_SCOPE = 2, 'ic', 'indicators'
+    IND_CT_SCOPE = 2, 'ic', 'indicator constraints'
     QUAD_CT_SCOPE = 3, 'qc', 'quadratic constraints'
     PWL_CT_SCOPE = 4, 'pwl', 'piecewise constraints'
     SOS_SCOPE = 5, 'sos', 'SOS'
 
+    def is_constraint_scope(self):
+        return self._value_ in frozenset([1, 2, 3, 4])
+
 
 class QualityMetric(Enum):
     def __new__(cls, code, has_int, cpx_codename):
@@ -512,7 +522,7 @@ class BasisStatus(Enum):
     See Also:
         The list of possible values for basis status can be found in the CPLEX documentation:
 
-        https://www.ibm.com/support/knowledgecenter/SSSA5P_12.10.0/ilog.odms.cplex.help/refcallablelibrary/cpxapi/getbase.html
+        https://www.ibm.com/support/knowledgecenter/SSSA5P_20.1.0/ilog.odms.cplex.help/refcallablelibrary/cpxapi/getbase.html
 
     """
 
@@ -545,11 +555,11 @@ class WriteLevel(Enum):
 
     The possible values are (in order of decreasing quantity of information written).
 
-        - AllVars: all variables are written
-        - DiscreteVars: all discrete variables are  written (binary, integer, semi-integer)
-        - NonZeroVars: all non-zero vars are written, regardless of their type.
-        - DiscreteNonZeroVars: all discrete non-zero vars are written.
-        - Auto: automatic value, same as DiscreteVars.
+        - AllVars (1): all variables are written
+        - DiscreteVars (2): all discrete variables are  written (binary, integer, semi-integer)
+        - NonZeroVars (3): all non-zero vars are written, regardless of their type.
+        - DiscreteNonZeroVars (4): all discrete non-zero vars are written.
+        - Auto (0): automatic value, same as DiscreteVars.
 
     *New in version 2.10*
     """
@@ -570,7 +580,7 @@ class WriteLevel(Enum):
     def filter_zeros(self):
         return self in {WriteLevel.NonZeroVars, WriteLevel.NonZeroDiscreteVars}
 
-    def filter_discrete(self):
+    def filter_nondiscrete(self):
         return self in {WriteLevel.Auto, WriteLevel.DiscreteVars, WriteLevel.NonZeroDiscreteVars}
 
     @classmethod
@@ -598,7 +608,7 @@ class EffortLevel(Enum):
     See Also:
         The list of possible values for effort level status can be found in the CPLEX documentation:
 
-https://www.ibm.com/support/knowledgecenter/SSSA5P_12.10.0/ilog.odms.cplex.help/refcppcplex/html/enumerations/IloCplex_MIPStartEffort.html
+https://www.ibm.com/support/knowledgecenter/SSSA5P_20.1.0/ilog.odms.cplex.help/refcppcplex/html/enumerations/IloCplex_MIPStartEffort.html
 
 
     """
