@@ -1,13 +1,13 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
 
 from docplex.mp.solution import SolveSolution
 from docplex.mp.sdetails import SolveDetails
-from docplex.mp.utils import DOcplexException
+from docplex.mp.utils import DOcplexException, is_number
 from docplex.mp.error_handler import docplex_fatal
 from docplex.util.status import JobSolveStatus
 from docplex.mp.constants import CplexScope
@@ -24,7 +24,7 @@ class IEngine(object):
     def register_callback(self, cb):
         raise NotImplementedError  # pragma: no cover
 
-    def connect_progress_listeners(self, listeners, model):
+    def connect_progress_listeners(self, model, listeners, qlisteners):
         raise NotImplementedError  # pragma: no cover
 
     def solve(self, mdl, parameters, **kwargs):
@@ -68,7 +68,7 @@ class IEngine(object):
     def get_solve_status(self):
         """  Return a DOcplexcloud-style solve status.
 
-        Possible enums are in docloud/status.py
+        Possible enums are in JobSolveStatus
         Default is UNKNOWN at this stage. Redefined for CPLEX and DOcplexcloud engines.
         """
         return JobSolveStatus.UNKNOWN  # pragma: no cover
@@ -229,7 +229,7 @@ class MinimalEngine(IEngine):
         # no callbacks
         pass  # pragma: no cover
 
-    def connect_progress_listeners(self, listeners, model):
+    def connect_progress_listeners(self, model, listeners, qlisteners):
         # no listeners
         if listeners:
             self.only_cplex(mname="connect_progress_listeners")  # pragma: no cover
@@ -424,7 +424,7 @@ class DummyEngine(IEngine):
     def unregister_callback(self, cb):
         pass  # pragma: no cover
 
-    def connect_progress_listeners(self, listeners, model):
+    def connect_progress_listeners(self, model, listeners, qlisteners):
         if listeners:
             model.warning("Progress listeners require CPLEX, not supported on engine {0}.".format(self.name))
 
@@ -502,48 +502,45 @@ class IndexerEngine(DummyEngine):
     An abstract engine facade which generates unique indices for variables, constraints
     """
 
-    def __init__(self, initial_index=0):
+    def __init__(self):
         DummyEngine.__init__(self)
-        self._initial_index = initial_index  # CPLEX indices start at 0, not 1
-        self.__var_counter = self._initial_index
-        self._ct_counter = self._initial_index
+        self._counters = {sc: 0 for sc in CplexScope}
 
-    def _increment_vars(self, size):
-        self.__var_counter += size
-        return self.__var_counter
-
-    def _increment_cts(self, size):
-        self._ct_counter += size
-        return self._ct_counter
-
-    def create_one_variable(self, vartype, lb, ub, name):
-        old_count = self.__var_counter
-        self._increment_vars(1)
+    def _scope_incr1(self, sc):
+        old_count = self._counters[sc]
+        self._counters[sc] += 1
         return old_count
 
+    def _scope_incr(self, sc, size):
+        assert size >= 1
+        old_count = self._counters[sc]
+        new_count = old_count + size
+        self._counters[sc] = new_count
+        return range(old_count, new_count)
+
+    def create_one_variable(self, vartype, lb, ub, name):
+        return self._scope_incr1(CplexScope.VAR_SCOPE)
+
     def create_variables(self, nb_vars, vartype, lb, ub, name):
-        old_count = self.__var_counter
-        new_count = self._increment_vars(nb_vars)
-        return list(range(old_count, new_count))
+        return self._scope_incr(CplexScope.VAR_SCOPE, nb_vars)
 
     def create_multitype_variables(self, keys, vartypes, lbs, ubs, names):
-        old_count = self.__var_counter
-        new_count = self._increment_vars(len(keys))
-        return list(range(old_count, new_count))
+        if is_number(keys):
+            l_keys = keys
+            assert l_keys >= 0
+        else:
+            l_keys = len(keys)
+        return self._scope_incr(CplexScope.VAR_SCOPE, l_keys)
 
     def _create_one_ct(self):
-        old_ct_count = self._ct_counter
-        self._increment_cts(1)
-        return old_ct_count
+        return self._scope_incr1(CplexScope.LINEAR_CT_SCOPE)
 
     def create_linear_constraint(self, binaryct):
         return self._create_one_ct()
 
     def create_batch_cts(self, ct_seq):
-        old_ct_count = self._ct_counter
         size = sum(1 for _ in ct_seq)  # iterator is consumed
-        self._increment_cts(size)
-        return range(old_ct_count, self._ct_counter)
+        return self._scope_incr(CplexScope.LINEAR_CT_SCOPE, size)
 
     def create_block_linear_constraints(self, ct_seq):
         return self.create_batch_cts(ct_seq)
@@ -552,16 +549,17 @@ class IndexerEngine(DummyEngine):
         return self._create_one_ct()
 
     def create_logical_constraint(self, logct, is_equivalence):
-        return self._create_one_ct()
+        return self._scope_incr1(CplexScope.IND_CT_SCOPE)
 
     def create_batch_logical_constraints(self, logcts, is_equivalence):
-        return self.create_batch_cts(logcts)
+        size = sum(1 for _ in logcts)  # iterator is consumed
+        return self._scope_incr(CplexScope.IND_CT_SCOPE, size)
 
     def create_quadratic_constraint(self, ind):
-        return self._create_one_ct()
+        return self._scope_incr1(CplexScope.QUAD_CT_SCOPE)
 
     def create_pwl_constraint(self, pwl_ct):
-        return self._create_one_ct()
+        return self._scope_incr1(CplexScope.PWL_CT_SCOPE)
 
     def get_all_reduced_costs(self, mdl):
         return {}
@@ -597,6 +595,9 @@ class IndexerEngine(DummyEngine):
         """
         return parameter.get()
 
+    def create_sos(self, sos):
+        return self._scope_incr1(CplexScope.SOS_SCOPE)
+
 
 class NoSolveEngine(IndexerEngine):
 
@@ -618,7 +619,7 @@ class NoSolveEngine(IndexerEngine):
 
     @staticmethod
     def _no_cplex_error(mdl, method_name):  # pragma: no cover
-        mdl.fatal("No CPLEX DLL and no DOcplexcloud credentials: {0} is not available".format(method_name))
+        mdl.fatal("No CPLEX runtime found: {0} is not available".format(method_name))
 
     def solve(self, mdl, parameters, **kwargs):  # pragma: no cover
         """
@@ -639,8 +640,11 @@ class NoSolveEngine(IndexerEngine):
     def make_from_model(mdl):
         # used in pickle
         eng = NoSolveEngine(mdl)
-        eng._increment_vars(mdl.number_of_variables)
-        eng._increment_cts(mdl.number_of_constraints)
+        eng._scope_incr(CplexScope.VAR_SCOPE, mdl.number_of_variables)
+        eng._scope_incr(CplexScope.LINEAR_CT_SCOPE, mdl.number_of_linear_constraints)
+        eng._scope_incr(CplexScope.IND_CT_SCOPE, mdl.number_of_logical_constraints)
+        eng._scope_incr(CplexScope.QUAD_CT_SCOPE, mdl.number_of_quadratic_constraints)
+        # TODO: add other scopes
         return eng
 
 
@@ -654,6 +658,7 @@ class ZeroSolveEngine(IndexerEngine):
     def __init__(self, mdl, **kwargs):
         IndexerEngine.__init__(self)  # pragma: no cover
         self._last_solved_parameters = None
+        self._mdl = mdl
 
     def show_parameters(self, params):
         if params is None:
@@ -677,6 +682,12 @@ class ZeroSolveEngine(IndexerEngine):
     def get_var_zero_solution(dvar):
         return max(0, dvar.lb)
 
+    def _create_parameter_sets(self, mdl):
+        self._mdl.fatal("Missing CPLEX engine to create parameterset")
+
+    def _build_multiobj_paramsets(self, mdl, lex_timelimits, lex_mipgaps):
+        self._mdl.fatal("Missing CPLEX engine to create parameterset")
+
     def solve(self, mdl, parameters, **kwargs):
         # remember last solved params
         self._last_solved_parameters = parameters.clone() if parameters is not None else None
