@@ -1,10 +1,9 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
-from __future__ import print_function
 from docplex.mp.utils import DOcplexException, DOcplexLimitsExceeded, resolve_pattern, is_int, is_string
 
 from enum import Enum
@@ -210,9 +209,9 @@ class AbstractErrorHandler(IErrorHandler):
         docplex_error_stop_here()
         raise DOcplexException(resolved_message)
 
-    def fatal_limits_exceeded(self):
+    def fatal_limits_exceeded(self, nb_vars, nb_constraints):
         docplex_error_stop_here()
-        raise DOcplexLimitsExceeded()
+        raise DOcplexLimitsExceeded(nb_vars, nb_constraints)
 
 
     def ok(self):
@@ -247,7 +246,7 @@ def docplex_error_stop_here():
     pass
 
 
-def docplex_add_trivial_infeasible_ct(ct):
+def docplex_add_trivial_infeasible_ct_here():
     # INTERNAL: set breakpoint here to inspect ct
     pass
 
@@ -256,10 +255,12 @@ def docplex_fatal(msg, *args):
     docplex_error_stop_here()
     raise DOcplexException(resolved_message)
 
-is_debug = os.environ.get('DOCPLEX_DEBUG')
+def is_docplex_debug():
+    return not not os.environ.get('DOCPLEX_DEBUG')
+
 
 def docplex_debug_msg(*args):
-    if is_debug:
+    if is_docplex_debug():
         msg = ' '.join(str(x) for x in args)
         #print(f"-- {msg}")
         print("-- {0}".format(msg))
@@ -289,3 +290,15 @@ class SilentErrorHandler(AbstractErrorHandler):
         pass
 
 
+def handle_error(logger, error, msg):
+    if error == "raise":
+        raise logger.fatal(msg)
+    elif error == "warn":
+        logger.warn(msg)
+    elif error =="ignore":
+        pass
+    else:
+        print("unknown error directive: {0}, expecting ignore|warn|raise, msg is: {1}".format(error, msg))
+
+
+
