@@ -1,13 +1,13 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
 # gendoc: ignore
 
 import re
-import six
+from docplex.mp.utils import is_string
 
 
 class ExchangeFormat(object):
@@ -71,15 +71,14 @@ class ExchangeFormat(object):
     def fromstring(definition):
         if definition:
             definition = definition.lower()
-        return _FORMAT_MAPPER.get(definition, None)
+        return _xtension2fmt_dict.get(definition, None)
 
 
 class LPFormat(ExchangeFormat):
-
     __raw = " -+/\\<>"
     __cooked = "_mp____"
 
-    from docplex.mp.compat23 import mktrans
+    from docplex.mp.utils import mktrans
 
     _str_translate_table = mktrans(__raw, __cooked)
     _unicode_translate_table = {}
@@ -100,15 +99,12 @@ class LPFormat(ExchangeFormat):
         return raw_name.translate(cls._unicode_translate_table)
 
     # which translate_method to use
-    if six.PY2:
-        _translate_chars = _translate_chars_py2
-    else:  # pragma: no cover
-        _translate_chars = _translate_chars_py3
+    _translate_chars = _translate_chars_py3
 
     def __init__(self):
         ExchangeFormat.__init__(self, "LP", "lp", requires_cplex=False)
 
-    lp_re = re.compile(r"[a-df-zA-DF-Z!#$%&()/,;?@_`'{}|\"][a-zA-Z0-9!#$%&()/.,;?@_`'{}|\"]*")
+    lp_re = re.compile(r"[a-df-zA-DF-Z!#$%&()/,;?@_`'{}|\"][a-zA-Z0-9!#$%&~()/.,;?@_`'{}|\"]*")
 
     @classmethod
     def is_lp_compliant(cls, name, lp_re=lp_re):
@@ -166,9 +162,39 @@ class LPFormat(ExchangeFormat):
 # noinspection PyPep8
 LP_format  = LPFormat()
 SAV_format = ExchangeFormat("SAV", "sav", requires_cplex=True, is_binary=True)
+SAVgz_format = ExchangeFormat("SAVgz", "sav.gz", requires_cplex=True, is_binary=True)
 OPL_format = ExchangeFormat("OPL", ".mod", requires_cplex=False)
 MPS_format = ExchangeFormat("MPS", ".mps", requires_cplex=True)
 
-_FORMAT_MAPPER = {"lp": LP_format,
-                  "opl": OPL_format
-                  }
+_xtension2fmt_dict = {"lp": LP_format,
+                      "opl": OPL_format,
+                      "mod": OPL_format,
+                      "mps": MPS_format,
+                      "sav": SAV_format,
+                      "sav.gz": SAVgz_format,
+                      "savgz": SAVgz_format
+                      }
+
+
+def parse_format(format_arg, error="raise"):
+
+    def fmt_error():
+        expected = "lp|mps|sav|sav.gz" #  "|".join(sorted(fk for fk in iterkeys(_xtension2fmt_dict)))
+        if error == "raise":
+            raise ValueError("unexpected format spec: {0}, possible: {1}".format(format_arg, expected))
+        elif error == "warn":
+            print("unexpected format spec: {0}, possible: {1}".format(format_arg, expected))
+        else:
+            pass
+
+    fmt = None
+    if isinstance(format_arg, ExchangeFormat):
+        fmt = format_arg
+    elif is_string(format_arg):
+        fmt_key = format_arg[1:] if format_arg.startswith(".") else format_arg
+        fmt = _xtension2fmt_dict.get(fmt_key.lower())
+        if fmt is None:
+            fmt_error()
+    else:
+        fmt_error()
+    return fmt
