@@ -1,7 +1,7 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
 # gendoc: ignore
@@ -10,24 +10,24 @@ from itertools import product, islice
 
 from docplex.mp.sosvarset import SOSVariableSet
 from docplex.mp.operand import LinearOperand
-from docplex.mp.dvar import Var
+from docplex.mp.dvar import Var, is_var
 from docplex.mp.linear import MonomialExpr, LinearExpr, AbstractLinearExpr, ZeroExpr, ConstantExpr
 from docplex.mp.operand import Operand
-from docplex.mp.constants import ComparisonType, UpdateEvent, ObjectiveSense
+from docplex.mp.constants import ComparisonType, UpdateEvent
 from docplex.mp.constr import LinearConstraint, RangeConstraint, \
     IndicatorConstraint, PwlConstraint, EquivalenceConstraint, IfThenConstraint, NotEqualConstraint, QuadraticConstraint
 from docplex.mp.functional import MaximumExpr, MinimumExpr, AbsExpr, PwlExpr, LogicalAndExpr, LogicalOrExpr, \
     LogicalNotExpr
 from docplex.mp.pwl import PwlFunction
-from docplex.mp.compat23 import fast_range, izip_longest
+
+from itertools import zip_longest as izip_longest
 from docplex.mp.environment import Environment
 from docplex.mp.utils import DOcplexException, DocplexQuadToLinearException
-from docplex.mp.utils import MultiObjective
+from docplex.mp.utils import MultiObjective, MockIterable
 from docplex.mp.utils import is_string, is_pandas_dataframe, is_function, is_iterable, is_int, is_number, has_len,\
-                              is_iterator, str_maxed, generate_constant, is_ordered_sequence
+         is_iterator, str_maxed, generate_constant, is_ordered_sequence
 
 
-from docplex.mp.compat23 import izip
 from docplex.mp.kpi import KPI
 from docplex.mp.solution import SolveSolution
 from docplex.mp.sttck import StaticTypeChecker
@@ -84,7 +84,7 @@ def fix_format_string(fmt, dimen=1, key_format='_%s'):
 def is_tuple_w_standard_str(z):
     if isinstance(z, tuple):
         zclass = z.__class__
-        return zclass is tuple or not ("__str__" in zclass.__dict__)
+        return zclass is tuple or "__str__" not in zclass.__dict__
     else:
         return False
 
@@ -132,7 +132,7 @@ def compile_naming_function(keys, user_name, dimension=1, key_format=None,
         if len(list_names) < len(keys):
             raise DOcplexException("An array of names should have same len as keys, expecting: {0}, go: {1}"
                                    .format(len(keys), len(list_names)))
-        key_to_names_dict = {k: nm for k, nm in izip(keys, list_names)}
+        key_to_names_dict = {k: nm for k, nm in zip(keys, list_names)}
         # use a closure
         return lambda k: key_to_names_dict[k]  # if k in key_to_names_dict else default_fn()
 
@@ -142,8 +142,10 @@ def compile_naming_function(keys, user_name, dimension=1, key_format=None,
 
 
 class _AbstractModelFactory(object):
-    def __init__(self, model):
+
+    def __init__(self, model, engine):
         self._model = model
+        self._engine = engine
         self._checker = model._checker
         ordered = model.keep_ordering
         self.term_dict_type = self._term_dict_type_from_ordering(model.keep_ordering)
@@ -153,6 +155,10 @@ class _AbstractModelFactory(object):
             from collections import OrderedDict
             self._var_dict_type = OrderedDict
 
+    def update_engine(self, engine):
+        # the model has already disposed the old engine, if any
+        self._engine = engine
+
     def var_dict_type(self, ordered):
         return self._var_dict_type if ordered else dict
 
@@ -173,6 +179,7 @@ class _AbstractModelFactory(object):
 
 
 class ModelFactory(_AbstractModelFactory):
+    status_var_prefix = '_bool'
     status_var_fmt = '_bool{{{0:s}}}'
 
     @staticmethod
@@ -180,11 +187,10 @@ class ModelFactory(_AbstractModelFactory):
         return default_bound if bound is None else float(bound)
 
     def __init__(self, model, engine):
-        _AbstractModelFactory.__init__(self, model)
+        _AbstractModelFactory.__init__(self, model, engine)
 
         self._var_container_counter = 0
         self.number_validation_fn = model._checker.get_number_validation_fn()
-        self._engine = engine
         self.stringifier = str_flatten_tuple
 
     @property
@@ -204,14 +210,10 @@ class ModelFactory(_AbstractModelFactory):
     def warning(self, msg, *args):
         self._model.warning(msg, args)
 
-    def update_engine(self, engine):
-        # the model has already disposed the old engine, if any
-        self._engine = engine
-
-    def _make_new_var(self, vartype, lb, ub, varname, origin=None):
+    def _make_new_var(self, vartype, lb, ub, varname, origin=None, var_factory_fn=Var):
         self_model = self._model
         idx = self._engine.create_one_variable(vartype, lb, ub, varname)
-        var = Var(self_model, vartype, varname, lb, ub, _safe_lb=True, _safe_ub=True)
+        var = var_factory_fn(self_model, vartype, varname, lb, ub, _safe_lb=True, _safe_ub=True)
         self_model._register_one_var(var, idx, varname)
         var.origin = origin
         return var
@@ -237,10 +239,15 @@ class ModelFactory(_AbstractModelFactory):
         if model.ignore_names or status_var_fmt is None:
             varname = None
         else:
-            # use name if any else use truncated ct string representation
-            base_varname = self.status_var_fmt.format(ct.name or str_maxed(ct, maxlen=20))
+            # use name if any
+            if ct.name:
+                base_varname = self.status_var_fmt.format(ct.name)
+            else:
+                status_index = model._new_ct_status_index()
+                base_varname = f"{self.status_var_prefix}#{status_index}"
+            # base_varname = self.status_var_fmt.format(ct.name or str_maxed(ct, maxlen=20))
             # if name is already taken, use unique index at end to disambiguate
-            varname = model._get_non_ambiguous_varname(base_varname)
+            varname = model._disambiguate_varname(base_varname)
 
         return self._make_new_var(binary_vartype, 0, 1, varname, origin=ct)
 
@@ -253,14 +260,13 @@ class ModelFactory(_AbstractModelFactory):
         if is_iterable(keys):
             if is_pandas_dataframe(keys):
                 used_keys = keys.index.values
-            elif has_len(keys):
-                used_keys = keys
-            elif is_iterator(keys):
-                used_keys = list(keys)
             else:
-                # TODO: make a test for this case.
-                self.fatal("Cannot handle iterable var keys: {0!s} : no len() and not an iterator",
-                           keys)  # pragma: no cover
+                try:
+                    used_keys = list(keys)
+                except:
+                    # TODO: make a test for this case.
+                    self.fatal("Cannot handle iterable var keys: {0!s} : no len() and not an iterator",
+                               keys)  # pragma: no cover
 
         elif is_int(keys) and keys >= 0:
             # if name is str and we have a size, disable automatic names
@@ -274,18 +280,30 @@ class ModelFactory(_AbstractModelFactory):
             self._checker.typecheck_key_seq(used_keys)
         return used_name, used_keys
 
+    max_flatten_tuple_size = 100
+
+    def _get_stringifier(self, dimension, keys):
+        if dimension > 1:
+            stringifier = str_flatten_tuple
+        elif len(keys) <= self.max_flatten_tuple_size:
+            stringifier = str_flatten_tuple
+        else:
+            first_key = next(iter(keys), None)
+            is_tuple = isinstance(first_key, tuple)
+            stringifier = str_flatten_tuple if is_tuple else str
+        return stringifier
+
     def _expand_names(self, keys, user_name, dimension, key_format):
         if user_name is None or self._model.ignore_names:
-            # no automatic names, ever
             return []
         else:
-            # default_naming_fn = self._model._create_automatic_varname
+            stringifier_ = self._get_stringifier(dimension, keys)
             actual_naming_fn = compile_naming_function(keys, user_name, dimension, key_format,
-                                                       stringifier=self.stringifier)
+                                                       stringifier=stringifier_)
             computed_names = [actual_naming_fn(key) for key in keys]
             return computed_names
 
-    def _check_bounds(self, nb_vars, bounds, default_bound, true_if_lb):
+    def _check_bound_seq(self, nb_vars, bounds, default_bound, true_if_lb):
         nb_bounds = len(bounds)
         bound_name = 'lb' if true_if_lb else 'ub'
         for b, b_value in enumerate(bounds):
@@ -315,17 +333,14 @@ class ModelFactory(_AbstractModelFactory):
 
         elif is_number(var_bound):
             self._checker.typecheck_num(var_bound, caller='in variable bound')
+            ignore_bound = False
             if true_if_lb:
-                if var_bound == default_bound:
-                    return []
-                else:
-                    return [float(var_bound)] * size
+                if default_bound is not None and var_bound == default_bound:
+                    ignore_bound = True
             else:
-                # ub
                 if var_bound >= default_bound:
-                    return []
-                else:
-                    return [float(var_bound)] * size
+                    ignore_bound = True
+            return [] if ignore_bound else [float(var_bound)] * size
 
         elif is_ordered_sequence(var_bound):
             nb_bounds = len(var_bound)
@@ -333,20 +348,19 @@ class ModelFactory(_AbstractModelFactory):
                 # see how we can use defaults for those missing bounds
                 self.fatal("Variable bounds list is too small, expecting: %d, got: %d" % (size, nb_bounds))
             else:
-                return self._check_bounds(size, var_bound, default_bound, true_if_lb)
+                return self._check_bound_seq(size, var_bound, default_bound, true_if_lb)
 
         elif is_iterator(var_bound):
-            # unfold the iterator, as CPLEX needs a list
-            return list(var_bound)
+            return self._check_bound_seq(size, list(var_bound), default_bound, true_if_lb)
 
         elif isinstance(var_bound, dict):
             dict_bounds = [var_bound.get(k, default_bound) for k in keys]
-            return self._check_bounds(size, dict_bounds, default_bound, true_if_lb)
+            return self._check_bound_seq(size, dict_bounds, default_bound, true_if_lb)
         else:
             # try a function?
             try:
                 fn_bounds = [var_bound(k) for k in keys]
-                return self._check_bounds(size, fn_bounds, default_bound, true_if_lb)
+                return self._check_bound_seq(size, fn_bounds, default_bound, true_if_lb)
 
             except TypeError:
                 self._bad_bounds_fatal(var_bound)
@@ -357,82 +371,145 @@ class ModelFactory(_AbstractModelFactory):
     def _bad_bounds_fatal(self, bad_bound):
         self.fatal("unexpected variable bound: {0!s}, expecting: None|number|function|iterable", bad_bound)
 
-    def new_multitype_var_list(self, size, vartypes, lbs=None, ubs=None, names=None):
+    # @staticmethod
+    # def safe_kth(array_or_empty, k, fallback=None):
+    #     if array_or_empty:
+    #         return array_or_empty[k]
+    #     else:
+    #         return fallback
+
+    def new_multitype_var_list(self, size, vartypes, lbs=None, ubs=None, names=None,
+                               var_factory_fn=Var):
+        # INTERNAL: called from model reader
         if not size:
             return []
         mdl = self._model
+
+        # -------------------------
         assert size == len(vartypes)
-        assert size == len(lbs)
-        assert size == len(ubs)
-        assert size == len(names)
+        assert not lbs or size == len(lbs)
+        assert not ubs or size == len(ubs)
+        assert not names or size == len(names)
+        # -------------------------
 
-        allvars = [Var(mdl, vartypes[k],
-                       names[k] if names[k] else None,
-                       lbs[k],
-                       ubs[k],
-                       _safe_lb=True,
-                       _safe_ub=True) for k in fast_range(size)]
+        if not lbs:
+            for v, vt in enumerate(vartypes):
+                if vt.is_semi_type():
+                    self.fatal("In position #{0}, you must provide an explicit lower bound for variable of type {1}"
+                               .format(v, vt.short_name))
 
-        indices = self._engine.create_multitype_variables(size, vartypes, lbs, ubs, names)
+        def safe_kth_functional(x_list, k, fallback_fn):
+            if x_list:
+                return x_list[k]
+            else:
+                return fallback_fn()
+
+        if mdl.ignore_names or not names:
+            names = MockIterable.from_value(size=size, val=None)
+
+        def kth_default_lb(k):
+            return vartypes[k].default_lb
+        def kth_default_ub(k):
+            return vartypes[k].default_ub
+        it_lbs = lbs or MockIterable(size, kth_default_lb)
+        it_ubs = ubs or MockIterable(size, kth_default_ub)
+
+        allvars = [var_factory_fn(mdl, vartypes[k],
+                                  names[k],
+                                  it_lbs[k],
+                                  it_ubs[k],
+                                  _safe_lb=True,
+                                  _safe_ub=True) for k in range(size)]
+        cpxnames = names or []  # no None
+        cpxlbs = lbs or []
+        cpxubs = ubs or []
+        indices = self._engine.create_multitype_variables(size, vartypes, cpxlbs, cpxubs, cpxnames)
         mdl._register_block_vars(allvars, indices, names)
         return allvars
 
-    def var_list(self, keys, vartype, lb, ub, name=None, key_format=None):
+    def var_list(self, keys, vartype, lb, ub, name=None, key_format=None, create_container=True):
         actual_name, fixed_keys = self.make_key_seq(keys, name)
-        ctn = self._new_var_container(vartype, key_list=[fixed_keys], lb=lb, ub=ub, name=name)
-        return self.new_var_list(ctn, fixed_keys, vartype, lb, ub, actual_name, 1, key_format)
+        if create_container:
+            ctn = self._new_var_container(vartype, key_list=[fixed_keys], lb=lb, ub=ub, name=name)
+            var_list = self.new_var_list(ctn, fixed_keys, vartype, lb, ub, actual_name, 1, key_format)
+            ctn._attach_var_list(var_list)
+        else:
+            var_list = self.new_var_list(None, fixed_keys, vartype, lb, ub, actual_name, 1, key_format, _add_container=False)
+        return var_list
 
     def new_var_list(self, var_container,
                      key_seq, vartype,
                      lb=None, ub=None,
                      name=str,
-                     dimension=1, key_format=None):
+                     dimension=1, key_format=None,
+                     _safe_bounds=False,
+                     _safe_names=False,
+                     var_factory_fn=Var,
+                     _add_container=True):
         number_of_vars = len(key_seq)
         if 0 == number_of_vars:
             return []
 
         # compute defaults once
-        default_lb = vartype.default_lb
+        default_lb = None if vartype.is_semi_type() else vartype.default_lb
         default_ub = vartype.default_ub
 
-        xlbs = self._expand_bounds(key_seq, lb, default_lb, number_of_vars, true_if_lb=True)
-        xubs = self._expand_bounds(key_seq, ub, default_ub, number_of_vars, true_if_lb=False)
+        if _safe_bounds:
+            assert not lb or len(lb) == number_of_vars
+            xlbs = lb or []
+            assert not ub or len(ub) == number_of_vars
+            xubs = ub or []
+        else:
+            xlbs = self._expand_bounds(key_seq, lb, default_lb, number_of_vars, true_if_lb=True)
+            xubs = self._expand_bounds(key_seq, ub, default_ub, number_of_vars, true_if_lb=False)
         # at this point both list are either [] or have size numberOfVars
 
-        all_names = self._expand_names(key_seq, name, dimension, key_format)
+        if _safe_names:
+            xnames = name or []
+        else:
+            xnames = self._expand_names(key_seq, name, dimension, key_format)
 
-        safe_lbs = not xlbs
-        safe_ubs = not xubs
-        if xlbs and xubs:
-            self._checker.check_vars_domain(xlbs, xubs, all_names)
+        safe_lbs = _safe_bounds or not xlbs
+        safe_ubs = _safe_bounds or not xubs
+        if xlbs and xubs and not _safe_bounds:
+            # no domain check from reader
+            self._checker.check_vars_domain(xlbs, xubs, xnames)
 
         mdl = self._model
-        allvars = [Var(mdl, vartype,
-                       all_names[k] if all_names else None,
-                       xlbs[k] if xlbs else default_lb,
-                       xubs[k] if xubs else default_ub,
-                       _safe_lb=safe_lbs,
-                       _safe_ub=safe_ubs) for k in fast_range(number_of_vars)]
+        itnames = xnames or MockIterable.from_value(size=number_of_vars, val=None)
+        itlbs = xlbs or MockIterable.from_value(number_of_vars, default_lb)
+        itubs = xubs or MockIterable.from_value(number_of_vars, default_ub)
+        allvars = [var_factory_fn(mdl,
+                                  vartype,
+                                  itnames[k],
+                                  itlbs[k],
+                                  itubs[k],
+                                  _safe_lb=safe_lbs,
+                                  _safe_ub=safe_ubs) for k in range(number_of_vars)]
 
         # query the engine for a list of indices.
-        indices = self._engine.create_variables(len(key_seq), vartype, xlbs, xubs, all_names)
-        mdl._register_block_vars(allvars, indices, all_names)
-        if var_container:
-            for dv in allvars:
-                mdl.set_var_container(dv, var_container)
+        indices = self._engine.create_variables(number_of_vars, vartype, xlbs, xubs, xnames)
+        mdl._register_block_vars(allvars, indices, xnames)
+        if _add_container:
+            # TRue if list is user-generated, False for dicts
+            mdl._add_var_container(var_container)
+
         return allvars
 
     def _make_var_dict(self, keys, var_list, ordered):
         self._checker.check_for_duplicate_keys(keys)
         _dict_type = self.var_dict_type(ordered)
-        vdict = _dict_type(izip(keys, var_list))
+        vdict = _dict_type(zip(keys, var_list))
         return vdict
 
     def new_var_dict(self, keys, vartype, lb, ub, name, key_format, ordered=False):
         actual_name, key_seq = self.make_key_seq(keys, name)
         ctn = self._new_var_container(vartype, key_list=[key_seq], lb=lb, ub=ub, name=name)
-        var_list = self.new_var_list(ctn, key_seq, vartype, lb, ub, actual_name, 1, key_format)
-        return self._make_var_dict(key_seq, var_list, ordered)
+        var_list = self.new_var_list(ctn, key_seq, vartype, lb, ub, actual_name, 1, key_format, _add_container=False)
+        var_dict = self._make_var_dict(key_seq, var_list, ordered)
+        self._model._add_var_container(ctn)
+        ctn._attach_var_dict(var_dict)
+        return var_dict
 
     def new_var_multidict(self, seq_of_key_seqs, vartype, lb, ub, name, key_format=None, ordered=False):
         # ---
@@ -452,7 +529,9 @@ class ModelFactory(_AbstractModelFactory):
         ctn = self._new_var_container(vartype, key_list=fixed_keys, lb=lb, ub=ub, name=name)
         cube_vars = self.new_var_list(ctn, all_key_tuples, vartype, lb, ub, name, dimension, key_format)
 
-        return self._make_var_dict(keys=all_key_tuples, var_list=cube_vars, ordered=ordered)
+        multidict = self._make_var_dict(keys=all_key_tuples, var_list=cube_vars, ordered=ordered)
+        ctn._attach_var_dict(multidict)
+        return multidict
 
     def new_var_df(self, keys1, keys2, vartype, lb=None, ub=None, name=None):  # pragma: no cover
         try:
@@ -467,7 +546,7 @@ class ModelFactory(_AbstractModelFactory):
         ctn = self._new_var_container(vartype, key_list=matrix_keys, lb=lb, ub=ub, name=name)
         lvars = self.new_var_list(ctn, matrix_keys, vartype, lb, ub, name, dimension=2, key_format=None)
         # TODO: see how to do without this temp dict
-        dd = dict(izip(matrix_keys, lvars))
+        dd = dict(zip(matrix_keys, lvars))
         # row-oriented dict
         rowd = {row_k: [dd[row_k, col_k] for col_k in col_keys] for row_k in row_keys}
         vdtf = DataFrame.from_dict(rowd, orient='index', dtype='object')
@@ -483,9 +562,6 @@ class ModelFactory(_AbstractModelFactory):
                 k = self_number_validation_fn(cst)
         return ConstantExpr(self._model, k)
 
-    def _new_safe_constant_expr(self, cst):
-        return ConstantExpr(self._model, cst)
-
     use_constant_expr = True
 
     def constant_expr(self, cst, safe_number=False):
@@ -494,8 +570,8 @@ class ModelFactory(_AbstractModelFactory):
         else:
             return self._new_constant_expr(cst, safe_number=safe_number)
 
-    def linear_expr(self, arg=None, constant=0, name=None, safe=False, transient=False):
-        return LinearExpr(self._model, arg, constant, name, safe=safe, transient=transient)
+    def linear_expr(self, arg=None, constant=0, safe=False, transient=False):
+        return LinearExpr(self._model, arg, constant, safe=safe, transient=transient)
 
     def _new_monomial_expr(self, dvar, coeff, safe=True):
         if coeff:
@@ -503,8 +579,6 @@ class ModelFactory(_AbstractModelFactory):
         else:
             return self.new_zero_expr()
 
-    _operand_types = (AbstractLinearExpr, Var, ZeroExpr)
-
     @staticmethod
     def _is_operand(arg, accept_numbers=False):
         return isinstance(arg, Operand) or (accept_numbers and is_number(arg))
@@ -527,6 +601,8 @@ class ModelFactory(_AbstractModelFactory):
                 used_msg = msg.format(e) if msg else qe.message
                 raise DOcplexException(used_msg)
 
+    _operand_types = (AbstractLinearExpr, Var, ZeroExpr)
+
     def _to_linear_expr(self, e, linexpr_class=LinearExpr, force_clone=False):
         # TODO: replace by to_linear_operand
         if isinstance(e, linexpr_class):
@@ -563,7 +639,7 @@ class ModelFactory(_AbstractModelFactory):
 
     def _new_binary_constraint(self, lhs, sense, rhs, name=None):
         # noinspection PyPep8
-        left_expr = self._to_linear_operand(lhs, msg="LinearConstraint. expects linear expressions, {0} was passed")
+        left_expr  = self._to_linear_operand(lhs, msg="LinearConstraint. expects linear expressions, {0} was passed")
         right_expr = self._to_linear_operand(rhs, msg="LinearConstraint. expects linear expressions, {0} was passed")
         self._checker.typecheck_two_in_model(self._model, left_expr, right_expr, "new_binary_constraint")
         ct = LinearConstraint(self._model, left_expr, sense, right_expr, name)
@@ -597,7 +673,8 @@ class ModelFactory(_AbstractModelFactory):
             self._model.error("infeasible range constraint, lb={0}, ub={1}, expr={2}", lb, ub, expr)
 
     def new_range_constraint(self, lb, expr, ub, name=None, check_feasible=True):
-        self._check_range_feasibility(lb, ub, expr)
+        if check_feasible:
+            self._check_range_feasibility(lb, ub, expr)
         linexpr = self._to_linear_operand(expr)
         rng = RangeConstraint(self._model, linexpr, lb, ub, name)
         linexpr.notify_used(rng)
@@ -615,10 +692,11 @@ class ModelFactory(_AbstractModelFactory):
 
     def new_if_then_constraint(self, if_ct, then_ct, negate=False):
         def check_bvar_eq_10(lhs, rhs):
-            return isinstance(lhs, Var) and\
-                   lhs.is_binary() and\
-                   rhs.is_constant() and\
-                   (rhs.get_constant() == 0 or rhs.get_constant() == 1)
+            if is_var(lhs) and lhs.is_binary():
+                as_const = rhs.as_constant()
+                return as_const == 0 or as_const == 1
+            else:
+                return None
 
         # INTERNAL
         m = self._model
@@ -647,11 +725,11 @@ class ModelFactory(_AbstractModelFactory):
 
     def new_batch_equivalence_constraints(self, bvars, linear_cts, active_values, names):
         return [self.new_equivalence_constraint(bv, lct, active, name)
-                for bv, lct, active, name in izip(bvars, linear_cts, active_values, names)]
+                for bv, lct, active, name in zip(bvars, linear_cts, active_values, names)]
 
     def new_batch_indicator_constraints(self, bvars, linear_cts, active_values, names):
         return [self.new_indicator_constraint(bv, lct, active, name)
-                for bv, lct, active, name in izip(bvars, linear_cts, active_values, names)]
+                for bv, lct, active, name in zip(bvars, linear_cts, active_values, names)]
 
     def new_binary_constraint_or(self, ct1, arg2):
         orexpr = self.new_logical_or_expr([ct1, arg2])
@@ -843,10 +921,8 @@ class ModelFactory(_AbstractModelFactory):
         pwl = PwlFunction(self_model, pwl_def=pwl_def, name=name)
         return pwl
 
-    def new_pwl_expr(self, pwl_func, e, usage_counter, y_var=None, add_counter_suffix=True, resolve=True):
+    def new_pwl_expr(self, pwl_func, e, y_var=None, resolve=True):
         return PwlExpr(self._model, pwl_func, e,
-                       usage_counter,
-                       add_counter_suffix=add_counter_suffix,
                        y_var=y_var,
                        resolve=resolve)
 
@@ -854,10 +930,6 @@ class ModelFactory(_AbstractModelFactory):
         self_model = self._model
         return PwlConstraint(self_model, pwl_expr, name)
 
-    @staticmethod
-    def default_objective_sense():
-        return ObjectiveSense.Minimize
-
     def new_kpi(self, kpi_arg, name_arg):
         # make a name
         if name_arg:
@@ -895,7 +967,7 @@ class ModelFactory(_AbstractModelFactory):
         posted_cts = []
         checker = self._checker
         filterfn = self._model._prepare_constraint
-        check_trivial = self._checker.check_trivial_constraints()
+        check_trivial = self._model._check_trivial_constraints()
         # look first
         ctseq = list(cts)
         if not ctseq:
@@ -943,17 +1015,17 @@ class ModelFactory(_AbstractModelFactory):
             names = generate_constant(None, None)
 
         ranges = [self.new_range_constraint(lb, exp, ub, name) for lb, exp, ub, name in
-                      izip(lbs, exprs, ubs, names)]
-        # else:
-        #     ranges = [self.new_range_constraint(lb, exp, ub) for lb, exp, ub in izip(lbs, exprs, ubs)]
+                  zip(lbs, exprs, ubs, names)]
         self._post_constraint_block(ranges)
         return ranges
 
     def new_solution(self, var_value_dict=None, name=None, objective_value=None, **kwargs):
         keep_zeros = kwargs.get('keep_zeros', True)
+        solved_by_arg = kwargs.get('solved_by')
 
         return SolveSolution(model=self._model, obj=objective_value,
                              var_value_map=var_value_dict, name=name,
+                             solved_by=solved_by_arg,
                              keep_zeros=keep_zeros)
 
     def _new_var_container(self, vartype, key_list, lb, ub, name):
@@ -963,7 +1035,6 @@ class ModelFactory(_AbstractModelFactory):
         ctn._index = old_varctn_counter
         ctn._index_offset = self._model.number_of_variables  # nb of variables before ctn
         self._var_container_counter = old_varctn_counter + 1
-
         return ctn
 
 
@@ -977,6 +1048,8 @@ class _VariableContainer(object):
         self._ub = ub
         self._name = name
         self._name_str = None
+        self._var_collection = []
+        self._iter_vars = iter([])
 
     @property
     def index(self):
@@ -1074,12 +1147,13 @@ class _VariableContainer(object):
         else:
             return next(islice(product(*self._keyss), relative_offset, None), None)
 
+    @property
     def shape(self):
         return tuple(len(k) for k in self._keyss)
 
     @property
     def dimension_string(self):
-        dim_string = "".join(["[%d]" % s for s in self.shape()])
+        dim_string = "".join(["[%d]" % s for s in self.shape])
         return dim_string
 
     def to_string(self):
@@ -1090,3 +1164,18 @@ class _VariableContainer(object):
 
     def __str__(self):
         return self.to_string()
+
+    def _attach_var_list(self, var_list):
+        self._var_collection = var_list
+        self._iter_vars = iter(self._var_collection)
+
+    def _attach_var_dict(self, var_dict):
+        self._var_collection = var_dict
+
+        def iter_dict_values(dd):
+            #  call of iter() is necessary here!
+            return iter(dd.values())
+        self._iter_vars = iter_dict_values(var_dict)
+
+    def __iter__(self):
+        return self._iter_vars
