@@ -1,26 +1,20 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
 
 # gendoc: ignore
-import logging
 import os
-import sched
 import tempfile
-import threading
-import time
 import sys
-
 from itertools import chain, repeat
-from six import PY2 as SIX_PY2
-from six import itervalues, iteritems
-
-
-from docplex.mp.compat23 import Queue, izip
 
+try:
+    from collections.abc import Callable
+except ImportError:
+    from collections import Callable  # @UnresolvedImport
 
 __int_types = {int}
 __float_types = {float}
@@ -28,24 +22,26 @@ __numpy_ndslot_type = None
 __numpy_matrix_type = None
 __pandas_series_type = None
 __pandas_dataframe_type = None
-__spark_dataframe_type = None
 
-try:
-    # noinspection PyUnresolvedReferences
-    type(long)  # @UndefinedVariable
-    # long is indeed a type we are in Python2,
-    __int_types.add(long)  # @UndefinedVariable
-except NameError:  # pragma: no cover
-    # long is not a type, do nothing
-    pass  # pragma: no cover
+# try:
+#     # noinspection PyUnresolvedReferences
+#     type(long)  # @UndefinedVariable
+#     # long is indeed a type we are in Python2,
+#     __int_types.add(long)  # @UndefinedVariable
+# except NameError:  # pragma: no cover
+#     # long is not a type, do nothing
+#     pass  # pragma: no cover
 
 try:
     import numpy
 
+    npv = numpy.__version__
+
     _numpy_is_available = True
 
     __int_types.add(numpy.bool_)
-    __int_types.add(numpy.bool)
+    if npv < '1.20':
+        __int_types.add(numpy.bool)
 
     __int_types.add(numpy.int_)
     __int_types.add(numpy.intc)
@@ -80,24 +76,10 @@ try:
 
     __pandas_series_type = Series
     __pandas_dataframe_type = DataFrame
-except ImportError:
+except ImportError:  # pragma: no cover
     __pandas_series_type = None
     __pandas_dataframe_type = None
 
-# 'findspark' must be executed if running in Windows environment, before importing Spark
-try:
-    import findspark  # @UnresolvedImport
-    findspark.init()
-except (ImportError, IndexError, ValueError):
-    pass
-
-try:
-    import pyspark
-
-    __spark_dataframe_type = pyspark.sql.dataframe.DataFrame
-except ImportError:
-    __spark_dataframe_type = None
-
 __int_types = frozenset(__int_types)
 
 
@@ -156,7 +138,8 @@ if _numpy_is_available:
             return retval
         except AttributeError:  # if s is not a numpy type, s.dtype triggers this
             return False
-else:
+
+else:  # pragma: no cover
     def numpy_is_numeric(t):
         return False
 
@@ -174,16 +157,6 @@ def is_number(s):
     return type_of_s in __all_python_num_types or numpy_is_numeric(type_of_s) or _is_numpy_ndslot(s)
 
 
-# def is_number2(s):
-#     if isinstance(s, __all_num_types_tuple):
-#         return True
-#     elif _numpy_is_available:
-#         type_of_s = type(s)
-#         if (numpy_is_numeric(type_of_s) or _is_numpy_ndslot(s)):
-#             return True
-#     else:
-#         return False
-
 def is_pandas_series(s):
     return __pandas_series_type is not None and type(s) is __pandas_series_type
 
@@ -200,19 +173,16 @@ def is_numpy_matrix(s):
     return __numpy_matrix_type and type(s) is __numpy_matrix_type
 
 
-def is_spark_dataframe(s):
-    return __spark_dataframe_type and isinstance(s, __spark_dataframe_type)
-
-
 string_types = {str}
-if SIX_PY2:
-    string_types.add(unicode)  # @UndefinedVariable
+# if SIX_PY2:
+#     string_types.add(unicode)  # @UndefinedVariable
 string_types = frozenset(string_types)
 
 
 def is_string(e):
     return type(e) in string_types
 
+
 def has_len(e):
     try:
         len(e)
@@ -248,11 +218,6 @@ def is_iterator(e):
 
 
 def is_function(e):
-    import platform
-    if platform.python_version() >= '3.7':
-        from collections.abc import Callable
-    else:
-        from collections import Callable  # @UnresolvedImport
     return isinstance(e, Callable)
 
 
@@ -271,18 +236,29 @@ def _to_list(arg):
 
 
 def _build_ordered_sequence_types():
-    if __pandas_series_type and __numpy_ndslot_type:
-        return (list, tuple, __pandas_series_type, __numpy_ndslot_type)
-    elif __pandas_series_type:
-        return (list, tuple, __pandas_series_type)
-    elif __numpy_ndslot_type:
-        return (list, tuple, __numpy_ndslot_type)
-    else:
-        return (list, tuple)
+    os_types = [list, tuple, range]
+
+    if __pandas_series_type:
+        os_types.append(__pandas_series_type)
+    if __numpy_ndslot_type:
+        os_types.append(__numpy_ndslot_type)
+    return tuple(os_types)
+
 
 def is_ordered_sequence1(arg, type_tuple=_build_ordered_sequence_types()):
     return isinstance(arg, type_tuple)
 
+
+def ordered_sequence_to_list(s, caller):
+    if is_pandas_series(s):
+        return s.tolist()
+    elif is_ordered_sequence(s):
+        return s
+    else:
+        raise DOcplexException('{0} requires ordered sequences: lists, numpy array or Series, got: {1}', caller,
+                               type(s))
+
+
 # def is_ordered_sequence2(arg):
 #     if isinstance(arg, (dict, str)):
 #         return False
@@ -326,10 +302,11 @@ class DOcplexException(Exception):
 
 
 class DOcplexLimitsExceeded(DOcplexException):
-    cplexce_limits_exceeded_msg = "**** Promotional version. Problem size limits exceeded, CPLEX code=1016"
+    cplexce_limits_exceeded_msg = \
+        "**** Promotional version. Problem size limits (1000 vars, 1000 consts) exceeded, model has {0} vars, {1} consts, CPLEX code=1016"
 
-    def __init__(self):
-        DOcplexException.__init__(self, self.cplexce_limits_exceeded_msg)
+    def __init__(self, nb_vars, nb_consts):
+        DOcplexException.__init__(self, self.cplexce_limits_exceeded_msg.format(nb_vars, nb_consts))
 
 
 class DocplexQuadToLinearException(DOcplexException):
@@ -347,6 +324,7 @@ class DocplexLinearRelaxationError(DOcplexException):
         msg = "Modeling object: {0!s} cannot be relaxed".format(obj)
         DOcplexException.__init__(self, msg)
 
+
 def normalize_basename(s, force_lowercase=True, maxlen=255):
     """Replaces some characters from s with a translation table:
     
@@ -368,8 +346,6 @@ def normalize_basename(s, force_lowercase=True, maxlen=255):
     """
     # replace all whietspaces by _
     l = s.lower() if force_lowercase else s
-    # table = mktrans(" ", "_")
-    # return l.translate(table)
     trans_table = {" ": "_",
                    "/": "_slash_",
                    "\\": "_backslash_",
@@ -387,9 +363,10 @@ def normalize_basename(s, force_lowercase=True, maxlen=255):
 
     if len(n) > maxlen - 8:
         h = format(hash(n) & 0xffffffff, "08x")
-        n = n[:maxlen-8] + "_"+ h
+        n = n[:maxlen - 8] + "_" + h
     return n
 
+
 def fix_whitespace(s, fix='_'):
     # replaces whitespaces by a character, default is '_'
     return s.replace(' ', fix)
@@ -411,21 +388,6 @@ def make_output_path2(actual_name, extension, basename_fmt, path=None):
     return os.path.join(output_dir, actual_basename)
 
 
-def make_path(error_handler, basename, extension, output_dir=None, name_transformer=None):
-    if output_dir is None:
-        output_dir = tempfile.gettempdir()
-    elif not os.path.exists(output_dir):
-        if not os.makedirs(output_dir):
-            error_handler.error("directory not found and not created: {0:s}", output_dir)
-            return None
-
-    norm_name = normalize_basename(basename)
-    basename = norm_name if not name_transformer else name_transformer % norm_name
-    filename = basename + extension
-    full_path = '/'.join([output_dir, filename])
-    return full_path
-
-
 def generate_constant(the_constant, count_max):
     if count_max is None:
         count_max = sys.maxsize
@@ -438,10 +400,6 @@ def generate_constant(the_constant, count_max):
 def iter_emptyset():
     return iter([])
 
-def iter_one(obj):
-    yield obj
-
-
 
 def resolve_pattern(pattern, args):
     """
@@ -463,7 +421,7 @@ def resolve_pattern(pattern, args):
         return pattern
 
 
-def str_maxed(arg, maxlen):
+def str_maxed(arg, maxlen, ellipsis_str=".."):
     """ Returns a (possibly) truncated string representation of arg
 
     If maxlen is positive (or null), returns str(arg) up to maxlen chars.
@@ -476,31 +434,14 @@ def str_maxed(arg, maxlen):
     if maxlen <= 0 or len(s) <= maxlen:
         return s
     else:
-        return "%s.." % s[:maxlen]
-
-
-try:
-    import scipy.sparse as sp
-except ImportError:
-    sp = None
-
-def is_scipy_sparse(m):
-    return sp and sp.issparse(m)
-
-
-def compute_is_index(seq, obj):
-    # assume obj is iterable multiple times
-    for i, elt in enumerate(seq):
-        if elt is obj:  # use 'is' to identify obj
-            return i
-
-    return None
+        return "%s%s" % (s[:maxlen], ellipsis_str)
 
 
 DOCPLEX_CONSOLE_HANDLER = None
 
 
 def get_logger(name, verbose=False):
+    import logging
     logging_level = logging.WARNING
     if verbose:
         logging_level = logging.DEBUG
@@ -536,205 +477,6 @@ def open_universal_newline(filename, mode):
             # for other errors, just raise them
             raise
 
-
-class CyclicLoop(object):
-    """ A cyclic loop executes actions at specified intervals, until
-    ``stop()`` is called.
-
-    This loop is based on sched.scheduler
-
-    Attributes:
-        stopped: True if the loop is stopped.
-    """
-
-    class Task(object):
-        """This class stores information needed to manage tasks.
-
-        Attributes:
-            id: The id of the task (automatically generated)
-            interval: The interval on which that task is called
-            action: The action function to call at ``interval``
-            argument: The arguments for the action function
-        """
-        id = 0
-        idgen_lock = threading.Lock()
-
-        def __init__(self, interval, priority, action, argument=()):
-            self.interval = interval
-            self.priority = priority
-            self.action = action
-            self.argument = argument
-            with self.idgen_lock:
-                self.id = CyclicLoop.Task.id
-                CyclicLoop.Task.id += 1
-
-    def __init__(self):
-        """Initialize a new empty CyclicLoop
-        """
-        self.stop_lock = threading.Lock()
-        self.stopped = False
-        self.scheduler = sched.scheduler(time.time, time.sleep)
-        # maps task id -> ev
-        self.events_by_id = {}
-        self.tasks_by_id = {}  # task id -> task
-
-    def enter(self, interval, priority, action, argument=()):
-        """Schedule a new event.
-
-        Works like sched.scheduler.enter(), but instead of a ``delay``, the
-        first argument is the ``interval`` the action must be performed.
-        """
-        with self.stop_lock:
-            if not self.stopped:
-                task = CyclicLoop.Task(interval, priority, action, argument)
-                self.tasks_by_id[task.id] = task
-                self._queue(task)
-
-    def _queue(self, task):
-        def process_task(task_id):
-            self._process_task(task_id)
-
-        ev = self.scheduler.enter(task.interval, task.priority, process_task, (task.id,))
-        self.events_by_id[task.id] = ev
-
-    def _process_task(self, task_id):
-        task = self.tasks_by_id[task_id]
-        task.action(*task.argument)
-        del self.events_by_id[task_id]
-        # do not reschedule if we are shutting down
-        with self.stop_lock:
-            if not self.stopped:
-                self._queue(task)
-
-    def start(self):
-        """Starts the loop. The loop stops only when ``stop()`` is called.
-        """
-        self.scheduler.run()
-
-    def stop(self):
-        """Stops the Loop.
-
-        When the loop is stopped, its ``stopped`` attribute is set immediately,
-        then all tasks in the scheduler are canceled.
-        """
-        with self.stop_lock:
-            self.stopped = True
-            for ev in itervalues(self.events_by_id):
-                try:
-                    self.scheduler.cancel(ev)
-                except ValueError:
-                    # if stop() is called from an event, the event has already
-                    # been triggered and poped'
-                    pass
-
-
-class ClosableQueue(Queue):
-    LAST = object()
-
-    def close(self):
-        self.put(self.LAST)
-
-    def __iter__(self):
-        while True:
-            item = self.get()
-            try:
-                if item is self.LAST:
-                    return
-                yield item
-            finally:
-                self.task_done()
-
-
-class ThreadedCyclicLoop(object):
-    """ A cyclic loop executes actions at specified intervals, until
-    ``stop()`` is called.
-
-    This loop is based on threads.
-
-    Attributes:
-        stopped: True if the loop is stopped.
-    """
-
-    class Task(threading.Thread):
-        """
-        Attributes:
-            id: The id of the task (automatically generated)
-            interval: The interval on which that task is called
-            action: The action function to call at ``interval``
-            argument: The arguments for the action function
-        """
-
-        def __init__(self, loop, interval, priority, action, argument=()):
-            super(ThreadedCyclicLoop.Task, self).__init__()
-            self.loop = loop
-            self.interval = interval
-            self.priority = priority
-            self.action = action
-            self.argument = argument
-            self.stopped = False
-
-        def run(self):
-            while not self.stopped:
-                # instead of one big sleep, do some smaller sleeps so that
-                # we can stop the thread with smaller granularity
-                for _ in range(self.interval):
-                    time.sleep(1)
-                    if self.stopped:
-                        break
-                if not self.stopped:
-                    self.perform()
-
-        def stop(self):
-            self.stopped = True
-
-        def perform(self):
-            self.action(*self.argument)
-
-    def __init__(self):
-        """Initialize a new empty ThreadedCyclicLoop
-        """
-        self.stop_lock = threading.Lock()
-        self.stopped = False
-        self.threads = set()
-        self.event_queue = ClosableQueue()
-
-    def enter(self, interval, priority, action, argument=()):
-        """Schedule a new event.
-
-        Works like sched.scheduler.enter(), but instead of a ``delay``, the
-        first argument is the ``interval`` the action must be performed.
-        """
-        with self.stop_lock:
-            if not self.stopped:
-                task = ThreadedCyclicLoop.Task(self, interval, priority,
-                                               action, argument)
-                self.threads.add(task)
-
-    def start(self, mt_worker=None, mt_arg=()):
-        """Starts the loop. The loop stops only when ``stop()`` is called.
-        """
-        for t in self.threads:
-            t.start()
-        if mt_worker:
-            while not self.stopped:
-                for task in self.event_queue:
-                    mt_worker(*((task,) + mt_arg))
-        for t in self.threads:
-            t.join()
-
-    def stop(self):
-        """Stops the Loop.
-
-        When the loop is stopped, its ``stopped`` attribute is set immediately,
-        then all tasks in the scheduler are canceled.
-        """
-        with self.stop_lock:
-            self.stopped = True
-        self.event_queue.close()
-        for t in self.threads:
-            t.stop()
-
-
 class _SymbolGenerator(object):
     def __init__(self, pattern, offset=1):
         ''' Initialize the counter and the pattern.
@@ -790,6 +532,7 @@ class _AutomaticSymbolGenerator(_SymbolGenerator):
         while True:
             yield self.new_symbol()
 
+
 class _IndexScope(object):
     # INTERNAL: full scope of indices.
 
@@ -804,8 +547,8 @@ class _IndexScope(object):
     def __getitem__(self, item):
         return self._index_map[item]
 
-    def count_filtered(self, filter):
-        return sum(1 for obj in self.iter_objects() if filter(obj))
+    def count_filtered(self, pred):
+        return sum(1 for obj in self.iter_objects() if pred(obj))
 
     @property
     def last(self):
@@ -813,18 +556,18 @@ class _IndexScope(object):
         if not size:
             raise ValueError("empty scope")
         else:
-            return self._index_map[size-1]
+            return self._index_map[size - 1]
 
     def iter_objects(self):
-        if SIX_PY2:
-            sorted_by_index = sorted((o for o in itervalues(self._index_map)), key=lambda o: o.index)
-            return iter(sorted_by_index)
-        else:
-            return iter(self._index_map.values())
+        # if SIX_PY2:
+        #     sorted_by_index = sorted((o for o in itervalues(self._index_map)), key=lambda o: o.index)
+        #     return iter(sorted_by_index)
+        # else:
+        return iter(self._index_map.values())
 
-    def generate_objects_filtered(self, filter):
+    def generate_objects_filtered(self, pred):
         for obj in self.iter_objects():
-            if filter(obj):
+            if pred(obj):
                 yield obj
 
     def generate_objects_with_type(self, obj_type):
@@ -853,7 +596,7 @@ class _IndexScope(object):
             # returns None if idx not in map
             return self._index_map.get(idx)
 
-    def reset(self):
+    def clear(self):
         self._index_map = {}
 
     def notify_obj_index(self, obj, index):
@@ -864,7 +607,7 @@ class _IndexScope(object):
         if indices:
             idxmap = self._index_map
             if idxmap is not None:
-                for obj, idx in izip(objs, indices):
+                for obj, idx in zip(objs, indices):
                     idxmap[idx] = obj
 
     # def update_indices(self):
@@ -886,10 +629,10 @@ class _IndexScope(object):
         nb_deleted = len(delset)
         idxmap = self._index_map
         if nb_deleted and idxmap is not None:
-            kept = [o for o in itervalues(idxmap) if o.index not in delset]
+            kept = [o for o in idxmap.values() if o.index not in delset]
             new_map = {}
             for nx, obj in enumerate(kept):
-                obj.index = nx
+                obj._set_index(nx)
                 new_map[nx] = obj
             self._index_map = new_map
 
@@ -897,14 +640,14 @@ class _IndexScope(object):
         if max_lines is None or max_lines < 0:
             max_lines = 1e+20
         print("-- scope: {0}[{1}]".format(self.qualifier, self.size))
-        for l, (k, o) in iteritems(self._index_map):
+        for l, (k, o) in self._index_map.items():
             if l > max_lines:
                 break
             print("  {0}> index={1}: {2:s}".format(l, k, o))
 
     def check_indices(self):
         # internal, checks that indices and keys are in sync
-        for k, o in iteritems(self._index_map):
+        for k, o in self._index_map.items():
             if k != o.index:
                 raise ValueError("key-index mismatch, key: {0}, index: {1}"
                                  .format(k, o.index))
@@ -954,8 +697,8 @@ def compute_overwrite_nb_threads(parameters, solver_context):
 
 def is_almost_equal(x1, x2, reltol, abstol=0):
     # returns true if x2 equals x1 w.r.t a miax of an absolute tolerance and a relative tolerance
-    prec = max(abstol, reltol* abs(x1))
-    return abs(x2-x1) <= prec
+    prec = max(abstol, reltol * abs(x1))
+    return abs(x2 - x1) <= prec
 
 
 class OutputStreamAdapter:
@@ -980,6 +723,11 @@ class OutputStreamAdapter:
             output_s = s.encode(self.encoding)
         self.stream.write(output_s)
 
+
+def mktrans(a, b):              # pragma: no cover
+    return str.maketrans(a, b)  # pragma: no cover
+
+
 def izip2_filled(it1, it2, **kwds):
     # izip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
     fillvalue = kwds.get('fillvalue')
@@ -996,7 +744,7 @@ def izip2_filled(it1, it2, **kwds):
     try:
         while iterators:
             tpl = tuple(map(next, iterators))
-            #print(tpl)
+            # print(tpl)
             yield tpl
     except _ZipExhausted:
         pass
@@ -1121,3 +869,95 @@ def is_quad_expr(obj):
         return obj.is_quad_expr()
     except AttributeError:
         return False
+
+
+def _var_match_function(source_model, target_model, match="auto", error='ignore'):
+    # returns a maping function from source vars
+    # to target model vars.
+    if match == "auto":
+        # if same stats, go for index:
+        if source_model.statistics == target_model.statistics:
+            match = 'index'
+        else:
+            match = 'name'
+    assert match != 'auto'
+    find_matching_var = None
+    if target_model is source_model:
+        def find_matching_var(dvar_, _):
+            return dvar_
+    elif match == "index":
+        def find_matching_var(dvar_, target_model_):
+            dvar = target_model_.get_var_by_index(dvar_.index)
+            return dvar
+    elif match == "name":
+        def find_matching_var(dvar1, target_model_):
+            key_name = dvar1.name
+            if key_name is None:
+                raise DOcplexException("Cannot find name matching variable for anonymous var, index: {0}")
+
+            dvar = target_model_.get_var_by_name(dvar1.name)
+            return dvar
+    else:
+        if callable(match):
+            # TODO: check signature of callable, expecting two args
+            def find_matching_var(dvar_, target_model_):
+                return match(dvar_, target_model_)
+        else:
+            source_model.fatal(
+                "Cannot interpret this as a variable match: {0!r} expecting auto|index|name or f(Var, Model)->Var",
+                match)
+
+    return find_matching_var
+
+
+
+
+class MockIterable:
+    """ Utility clas which behaves like a sequence of N items, with constant memory space
+
+    """
+
+    @classmethod
+    def from_value(cls, size, val):
+        """ Returns a mock iterable over a constant
+
+        :param size: size of the iterable
+        :param val:  constant value
+        :return: an instance of MockIterable
+        """
+        mock_iter_val = _MockIterableConstant(size, val)
+        return mock_iter_val
+
+    def __init__(self, size, fn_k):
+        self.fn_k = fn_k
+        self.size = size
+        assert size >= 0
+
+    def __len__(self):
+        return self.size
+
+    def _as_gen(self):
+        for i in range(self.size):
+            yield self.fn_k(i)
+
+    def __iter__(self):
+        return self._as_gen()
+
+    def __getitem__(self, index):
+        if 0 <= index < self.size:
+            return self.fn_k(index)
+        else:
+            raise IndexError
+
+class _MockIterableConstant(MockIterable):
+    def __init__(self, size, cst):
+        def fnk(k):
+            return cst
+        super().__init__(size, fnk)
+        self.size = size
+        self.constant = cst
+        assert size >= 0
+
+    def __getitem__(self, index):
+        return self.constant
+
