@@ -1,9 +1,10 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2022
 # --------------------------------------------------------------------------
 
+from docplex.mp.error_handler import docplex_fatal
 
 class VarType(object):
     """VarType()
@@ -20,19 +21,19 @@ class VarType(object):
         self._ub = ub
         self._cpx_typecode = cplex_typecode
 
-    # def get_cplex_typecode(self):
-    #     # INTERNAL
-    #     return self._cpx_typecode
+    def is_semi_type(self):
+        return False
 
     @property
     def cplex_typecode(self):
         """ This property returns the CPLEX type code for this type.
         Possible values are:
-            'B' for binary type
-            'I' for integer type
-            'C' for continuous type
-            'S' for semicontinuous type
-            'N' for semiinteger type
+        
+            - 'B' for binary type
+            - 'I' for integer type
+            - 'C' for continuous type
+            - 'S' for semicontinuous type
+            - 'N' for semiinteger type
 
         :return: a one-letter string.
         """
@@ -92,9 +93,9 @@ class VarType(object):
     @classmethod
     def _is_within_bounds_and_tolerance(cls, candidate_value, lb, ub, tolerance):
         assert tolerance >= 0
-        if candidate_value <= lb - tolerance:
+        if candidate_value < lb - tolerance:
             res = False
-        elif candidate_value >= ub + tolerance:
+        elif candidate_value > ub + tolerance:
             res = False
         else:
             res = True
@@ -248,6 +249,9 @@ class SemiContinuousVarType(VarType):
         VarType.__init__(self, short_name="semi-continuous", lb=1e-6, ub=plus_infinity, cplex_typecode='S')
         self._plus_infinity = plus_infinity
 
+    def is_semi_type(self):
+        return True
+
     def _compute_ub(self, candidate_ub, logger):
         return self._plus_infinity if candidate_ub >= self._plus_infinity else float(candidate_ub)
 
@@ -257,6 +261,11 @@ class SemiContinuousVarType(VarType):
                 'semi-continuous variable expects strict positive lower bound, not: {0}'.format(candidate_lb))
         return candidate_lb
 
+    @property
+    def default_lb(self):
+        # there is NO default lb
+        docplex_fatal("Type {0} has no default lower bound".format(self.short_name))
+
     def is_discrete(self):
         """ Checks if this is a discrete type.
 
@@ -281,11 +290,18 @@ class SemiIntegerVarType(VarType):
             This class models the :index:`semi-integer` variable type and
             is not meant to be instantiated.
     """
-
     def __init__(self, plus_infinity=1e+20):
         VarType.__init__(self, short_name="semi-integer", lb=1e-6, ub=plus_infinity, cplex_typecode='N')
         self._plus_infinity = plus_infinity
 
+    def is_semi_type(self):
+        return True
+
+    @property
+    def default_lb(self):
+        # there is NO default lb
+        docplex_fatal("Type {0} has no default lower bound".format(self.short_name))
+
     def _compute_ub(self, candidate_ub, logger):
         return min(candidate_ub, self._plus_infinity)
 
@@ -312,3 +328,7 @@ class SemiIntegerVarType(VarType):
 
     def __hash__(self):  # pragma: no cover
         return VarType._hash_vartype(self)
+
+
+
+
